#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <errno.h>


// GNU 
#include <DLList.h>


// PLinda
#include "C_Interface.h"
#include "plinda_daemon.h"
#include "ClientCommLink.h"
#include "Architectures.h"
#include "Timer.h"

// polling frequency
int minimum_idle_time = 30;
int idle_check_interval = 30;
int busy_check_interval = 2;
int poll_interval = 0;
int idle_flag = -1;
int msg_seq_num = 0;
int exit_flag = 0;

struct timeval idle_begin_time;
struct timeval idle_cur_time;

// worker info          ---------------------------------------------
class WorkerType {
public:
  // constructor.
  WorkerType(procid id,
             const char* name, 
	     int pid) {
    ::strcpy(x_workername,name); 
    x_identifier = id;
    x_pid = pid;
  }

  const char* name(void) { return x_workername; }
  procid identifier(void) { return x_identifier; }
  int UNIX_pid(void) { return x_pid; }

private:
  // data members.
  procid x_identifier;
  char x_workername[100];
  int x_pid;
};

typedef DLList<WorkerType*> WorkerListType;

WorkerListType worker_list;


// forward declaration  --------------------------------------------
int check_idleness(void);

int create_worker(procid, const char*, int);
WorkerType* remove_worker(int);
int terminate_workers(void);  // let defunct workers go...
int kill_workers(int);        // kill workers
int kill_all_workers(void);   // kill all the workers.

int handle_server_request(int);
int send_notice(int);


// global variables -------------------------------------------------

// hostname             
char hostname[100];

// info about this daemon.
procid my_pid;
int    my_temp_id;

// server address info;
unsigned short server_port;
const char* server_host = 0;
const char* x_server_host = 0;
const char *x_host_type = 0;
const char *x_server_host_type = 0;
int x_need_to_convert = 0;
int busy_interval = 0;
DaemonModes x_mode;

// local communication link.
ClientCommLink server_agent;



ifstream loggedFile;
long kbd_itime, mouse_itime, console_itime, cur_time, utmp_itime;
double one_load, five_load, fifteen_load;



int timerToSecs(struct timeval start, struct timeval end) 
{
  int secs,usecs;
  secs = end.tv_sec - start.tv_sec;
  usecs = end.tv_usec - start.tv_usec;  
  if(usecs > 500000)
    secs++;
  return secs;
}

void doMigrationTesting()
{
  if(busy_interval == 0) {
    send_notice(IDLE_NOTICE);
    cerr << "> " << hostname << " is idle forever\n";
    while(!exit_flag) {
      if(handle_server_request(poll_interval) == -1) 
	break;
      terminate_workers();    
    }
  }
  srand48(my_pid);
  while(!exit_flag) {
    if(handle_server_request(poll_interval) == -1) 
      break;
    terminate_workers();
    send_notice(IDLE_NOTICE);
    ::gettimeofday(&idle_begin_time, NULL);	  
    double idle_time = drand48() * busy_interval * 2;
    cerr << "> " << hostname << " is idle " << idle_time << endl << flush;
    sleep((int)idle_time);
    kill_all_workers();
    send_notice(BUSY_NOTICE);
    double busy_time = drand48() * busy_interval * 2;
    cerr << "> " << hostname << " is busy for " << busy_time << endl << flush;
    sleep((int)busy_time);
  }
}
int main(int argc, char** argv, char**) {

  // establish a connection to the server -------------------

  // argv[1]: process id.
  // argv[2]: server port.
  // argv[3]: server symbolic address.
  // argv[4]: server_host
  // argv[5]: host_type of this machine
  // argv[6]: host_type of the server
  // argv[7]: mode  = { MONITOR, MIGRATION, USE_FILE }
  // argv[8]: busy_interval if mode != MIGRATION this == 0

  assert(argc == 9);

  my_pid = procid(atol(argv[1]));
  server_port = (unsigned short)atoi(argv[2]);
  server_host = argv[3];
  x_server_host = argv[4];
  if(::strcmp(x_server_host,"?") == 0) x_server_host = 0;
  my_temp_id = ::getpid();
  x_host_type = argv[5];
  x_server_host_type = argv[6];
  x_mode = (DaemonModes)atoi(argv[7]);
  busy_interval = atoi(argv[8]);

  assert(busy_interval >= 0);  
  assert(x_mode == MONITOR || x_mode == MIGRATION || x_mode == USE_FILE);
  assert(x_mode == MIGRATION || busy_interval == 0);

  if(! PLinda_Architectures::supported(x_host_type)) {
    cout << "Daemon was not passed correct architecture type:" << x_host_type
	 << endl;
    assert(0);
  }
  if(! PLinda_Architectures::supported(x_server_host_type)) {
    cout << "Daemon was not passed correct architecture type:" << 
      x_server_host_type<< endl;
    assert(0);
  }
  x_need_to_convert = 
    PLinda_Architectures::need_to_convert(x_host_type,x_server_host_type);
   
  // initialize the local communication link.
  ::gethostname((char*)hostname, 100);
  server_port = ntohs(server_port);
  server_agent.initialize(my_pid, server_port, server_host, x_need_to_convert);
  // set up a temporary connection to the server.
  if(server_agent.request_link() == -1) {
    cerr << "> ***ERROR: daemon on " << hostname 
	 << " failed to set up a link to the server !\n";
    ::exit(1);
  }

  {
    Header msg;
    msg.daemonMsgType(REGISTER);
    msg.sequenceNum(msg_seq_num);
    msg.transientId(my_temp_id);
    if(server_agent.send(msg) == -1) {
      cerr << "***ERROR: daemon on " << hostname
	   << " failed to send a message to the server !\n";
      ::exit(1);
    }
  }

  if(x_mode == USE_FILE) {
    // open up the file!
    char fname[100];
    sprintf(fname,
	    "/merv.c/student/wyckoff/test_log_files/%s.uptime.log",hostname);
    cerr << "*********trying to open<"<<fname <<">" << endl <<flush;
    loggedFile.open(fname);
    if(!loggedFile) {
      cerr << "***********unable to open " << fname << endl << flush;
      ::exit(1);
    }
  }

  // main loop -------------------------------------------

  poll_interval = idle_check_interval;

  worker_list.clear();

  if(x_mode == MIGRATION) {
    doMigrationTesting();
    return 1;
  }
  
  while(!exit_flag) {
    sleep(1);
    if(check_idleness()) {
      if(idle_flag != 1) {
	cerr << "> " << hostname << " is idle !\n";
	cerr.flush();
	// set idle flag to be on.
	idle_flag = 1;
	// change the polling interval.
	poll_interval = busy_check_interval;	  
	// send idle notice to the server.
	send_notice(IDLE_NOTICE);
	::gettimeofday(&idle_begin_time, NULL);	  
      }
    } else { // !check_idleness()
      // kill all the processes.
      kill_all_workers();
      if(idle_flag != 0) {
	cerr << "> " << hostname << " is busy !\n";
	cerr.flush();
	// set idle flag to be off.
	idle_flag = 0;
	// change the polling interval.
	poll_interval = idle_check_interval;
	// send busy notice to the server.
	send_notice(BUSY_NOTICE);
      }
    }    
    // handle requests from the server, if any.
    if(handle_server_request(poll_interval) == -1) break;
    
    // check if there is any worker that has recently terminated.
    terminate_workers();

  } // while(1)
  
}



int handle_server_request(int timeout) {
  Header message;
  Tuple *tuple = NULL;
  int msg_flag;

  while(1) {
    msg_flag = server_agent.async_receive(message, tuple, timeout);
    if(msg_flag == -1) {
      cerr << "> daemon on " << hostname << 
	" : received an invalid message.\n";
      return -1;
    }
    
    if(msg_flag == 0) return 0;
    
    // interrrupted.
    if(msg_flag == -2) continue;
    
    int request = message.daemonMsgType();
    switch(request) {
    case TERMINATE:
      assert(tuple == NULL);
      kill_all_workers();
      exit_flag = 1;
      break;

    case CREATE_WORKER: 
      assert(tuple);
      if(idle_flag == 0) {
	Tuple::destroy(tuple);
	tuple = NULL;
	// busy, and so ignore a process creation request.
	break;
      }
      {
	int  interactive = message.interactive();
	procid identifier = message.processId();
	char fileName[100];
	tuple->read(0, TupleField::PLchar, fileName, 100);
	create_worker(identifier, fileName,interactive);
	Tuple::destroy(tuple);
	tuple = NULL;
      }
      break;
      
    case KILL_WORKER: 
      assert(tuple == NULL);
      {
	int pid = message.transientId();
	kill_workers(pid);
      }
      break;
      
    case SET_INTERVAL:
      assert(tuple == NULL);
      minimum_idle_time = message.idleTime();
      if(minimum_idle_time > 0) {
	poll_interval = busy_check_interval;
      } else {
	poll_interval = idle_check_interval;
      }
      cerr << "> daemon on " << hostname << 
	" : set minimum idle time to " << minimum_idle_time << ".\n";
      cerr.flush();
      break;
      
    case TEST_FAILURE:
      assert(tuple == NULL);
      {
	send_notice(ALIVE_NOTICE);
      }
      break;
      
    case RESET_TIMER:
      ::gettimeofday(&idle_begin_time, NULL);	  
      break;
    case GET_TIMER:
      send_notice(GET_TIMER);
      break;
    default:
      tuple = NULL;
      cerr << "> daemon on " << hostname << 
	" : received an invalid message.\n";
      kill_all_workers();
      return -1;
    }
    
    timeout = 0;
  }
  return 1;
}


int 
send_notice(int notice) {
  int res;

  Header message;
  message.daemonMsgType(notice);
  if(notice == BUSY_NOTICE || notice == GET_TIMER) {
    ::gettimeofday(&idle_cur_time, NULL);	  
    int idleSecs = timerToSecs(idle_begin_time,idle_cur_time);
    //    cerr <<"ilde secs = " << idleSecs << endl << flush;
    message.sequenceNum(idleSecs);
  }
  else
    message.sequenceNum(msg_seq_num);
  if(notice != ALIVE_NOTICE) {
    ++msg_seq_num;
  }

  assert(message.tupleLength() == 0);
  if((res=server_agent.send(message)) == -1) {
    cerr << "> daemon on " << hostname << " : failed to send a message.\n";
  }

  return res;
}


long mygettimeofday()
{

  if(x_mode == USE_FILE) {
    // read in the next line of the file
    if(loggedFile.eof()) {
      cerr << "That's it, ran out of data for the daemon on " << hostname << endl
	   << flush; 
      kill_all_workers();
      send_notice(BUSY_NOTICE);
      exit(1);
    }
    loggedFile >> cur_time >> kbd_itime >> mouse_itime >> 
      console_itime >> utmp_itime;
    int temp ;
    while( (temp = loggedFile.get()) == ' ') ; // just eat'm
    if(temp == '\n') {
      one_load = five_load = fifteen_load = 0;
    } else {
      loggedFile.putback(temp);
      loggedFile >> one_load >> five_load >> fifteen_load;
    }
    if(cur_time <= 0) 
      cerr << "****** daemon got a bad line!  failure!!!!\n" << flush;
    else
      cerr << "cur time = " << cur_time << "utmp_itime = " << utmp_itime <<
	" and fifteen load = " << fifteen_load << endl << flush;

    return cur_time;
  } else {
    struct timeval cur_time;
    gettimeofday(&cur_time, NULL);
    return cur_time.tv_sec;
  }
}

long mykbd_itime(long cur)
{
  if(x_mode == USE_FILE) {
    (void)cur;
    return kbd_itime;
  } else {
    struct stat stat_buf;
    ::stat("/dev/kbd", &stat_buf); 
    return cur - stat_buf.st_atime;
  }
}

long mymouse_itime(long cur)
{
  if(x_mode == USE_FILE) {
    (void)cur;
    return mouse_itime;
  } else {
    struct stat stat_buf;
    ::stat("/dev/mouse", &stat_buf);
    return cur - stat_buf.st_atime;
  }
}

long myconsole_itime(long cur)
{
  if(x_mode == USE_FILE) {
    (void)cur;
    return console_itime;
  } else {
    struct stat stat_buf;
    ::stat("/dev/console", &stat_buf);
    return cur - stat_buf.st_atime;
  }
}
// checkc idleness  -----------------------------------------------
int
check_idleness(void) {
  if(minimum_idle_time == 0) return 1;

  long cur_time =  mygettimeofday();
  long kbd_itime = mykbd_itime(cur_time);
  long mouse_itime = mymouse_itime(cur_time);
  long console_itime = myconsole_itime(cur_time);
  
  if(kbd_itime >= minimum_idle_time && mouse_itime >= minimum_idle_time &&
     console_itime >= minimum_idle_time)
    return 1;
  else
    return 0;

  //  return (idle_time >= minimum_idle_time);
}


// create a worker process.
int create_worker(procid identifier, const char* workername, int interactive) {
  int worker_pid;
  // create worker
  if((worker_pid = fork()) == 0) {
    // close the parent's connection.
    server_agent.close();

    if(interactive) {
      if(x_server_host == 0) {
	cerr << "> ***ERROR: interactive process but no x server host info.\n";
	cerr.flush();
	return 0;
      }

      char command[1000];
      ::sprintf(command,"plinda/lib/%s/%s %ld %d %s %s %d; cat",x_host_type,
		workername, identifier, server_port, server_host, 
		"interactive",x_need_to_convert);
      char xterm[100];
      ::sprintf(xterm,"plinda/bin/%s/xterm",x_host_type);
      if(::execl(xterm,"xterm","-display",x_server_host,"-e",
		 "/bin/sh", "-c", command, 0) == -1) {
	cerr << "> ****ERROR: daemon on " << hostname 
	     << " failed to create a process to run " << workername << "\n";
	cerr << "either plinda/lib/" << x_host_type << "/" << workername
	     << " or plinda/lib/" << x_host_type << "/xterm is not executable"
	     << endl;
	cerr << "command=<" << command << ">" << endl;
	cerr << "xterm=<" << xterm << ">\n";
	cerr.flush();
      }

    } else {
      char name[100];
      ::sprintf(name, "plinda/lib/%s/%s", x_host_type,workername);
      char cidentifier[100];
      ::sprintf(cidentifier, "%ld", identifier);
      char cport[100];
      ::sprintf(cport, "%d", server_port);
      char need_to_convert[10];
      ::sprintf(need_to_convert,"%d",x_need_to_convert);
      if(::execl(name, workername, cidentifier, cport, server_host, 
		 "noninteractive",need_to_convert,0) == -1) {
	cerr << "> *** ERROR: daemon on " << hostname 
	     << " failed to create a process to run " << workername << "\n";
	cerr << "plinda/lib/" << x_host_type << "/" << workername
	     << " does not seem to be executable" << endl;
	cerr.flush();
      }
    }
  }

  // register this new worker.
  WorkerType* worker = new WorkerType(identifier, workername, worker_pid);
  worker_list.prepend(worker);

  // report this.
  cerr << "> daemon on " << hostname << " created (" 
       << workername << "," << identifier << "," << worker_pid << ")\n";
  cerr.flush();

  return 1;
}


// kill the process which the server pointed to.

int kill_workers(int pid) {
  ::kill(pid, SIGKILL);
  return 1;
}


// this host becomes busy, and leave it alone to the owner 
// by killing all the workers immediately.

int kill_all_workers(void) {

  //  cerr << hostname << " is killing workers !!\n";

  for(Pix cur=worker_list.first(); cur != 0; worker_list.next(cur)) {
    WorkerType* worker = worker_list(cur);
    ::kill(worker->UNIX_pid(), SIGKILL);

    // cerr << hostname << " killed worker(" 
    //     << worker->identifier() << "," << worker->UNIX_pid() << ")\n"; 
     
  }

  cerr.flush();

  return worker_list.length();
}



// remove the entry for a worker process with given pid from the list.
WorkerType* remove_worker(int id) {
  Pix cur;
  for(cur=worker_list.first(); cur!=0; worker_list.next(cur)) {
    WorkerType* worker = (WorkerType*)worker_list(cur);
    if(worker->UNIX_pid() == id) {
      worker_list.del(cur);
      return worker;
    }
  }
  return 0;
}


// forked processes stay around as defunct 
// until 'wait' is explicitly called for them.
// here, we call wait for them and delete the corresponding entries.
int terminate_workers(void) {

  while(1) {
    int pid = ::wait3(0, WNOHANG, 0);

    if(pid <= 0) return 1;

    WorkerType* worker = remove_worker(pid);
    if(worker == 0) {
      cerr << "> *** ERROR: daemon on " << hostname << " can't find an entry "
	   << "for process " << pid << ".\n";
      cerr.flush();
    } else {
      cerr << "worker " << worker->identifier() << " on " 
           << hostname << " terminated.\n";
      cerr.flush();
    }

    delete worker;
  }

  return 1;
}



