// This may look like C code, but it is really -*- C++ -*-
//
// File:        plinda_lib.C 
// Description: the client library routines. 
// Created:      
// Author:      Karp Joo Jeong
// Mail:        jeong@cs.nyu.edu
// Note: 	The code is largely inspired by the client library  
//		that Brian E. Anderson wrote for PLinda 1.0.
//
// The PLinda pre-compiler "plc" converts PLinda language statements 
// to these run-time calls. Therefore, updates to these routines and 
// the pre-compiler should be coordinated so that they are always 
// consistent with each other.  
//     PL_proc_start
//     PL_proc_end
//     PL_create_group
//     PL_destroy_group
//     PL_xstart
//     PL_xcommit
//     PL_xabort
//     PL_reserve_physid
//     PL_out
//     PL_in
//     PL_rd
//     PL_physrdp
//     PL_physinp
//     PL_log_inout
//     PL_log_rdp
//     PL_rdp_args
//
// Changes to the number and type of arguments must be migrated to the
// preprocessor also.
//

#include <iostream.h>
//#include "plinda_lib.h"
#include "plinda_pragmas.h"
#include "plinda_requests.h"
#include "ErrorType.h"
#include "Machine.h"
#include "Header.h"
#include "ClientCommLink.h"
#include "C_Interface.h"
#include "Timer.h"

const int LARGE_TUPLE = 100000;

// the server agent -- bind to any port
static ClientCommLink server_agent;
static procid  my_pid 		= zero_procid;
static procid  my_temp_id 	= 0;

static bool inTransaction = false;
static bool needToStartTransaction = false;
// the die message.  if anything goes wrong the client dies reporting who
// he is and what the problem is.

char hostname[1000];
const char* execname;

DLList<Tuple*> theOutList;
DLList<Tuple*> replayList;

#define die_if(x) if(!x) die(x)
void die(const ErrorType& x) {
  cerr << "> error: process (" << execname << "," << my_pid << ") on " 
    << hostname << "\n";
  cerr << "         " << x.message() << ".\n";
  ::exit(1);
}


// our abort handler when we are in an evaled sub-process of a transaction
void client_die() {
  cerr << "Process (" << my_pid << ") caught an abort.\n";
  ::exit(1);
}

void client_kill() {
#ifdef PL_DEBUG
  cerr << "Process (" << my_pid << "( got a kill.\n";
#endif
  ::exit(-1);
}


Tuple *_PLtemp = NULL; // the tuple that is used by the client programs

// --------------------------------------------------------------------------
// forward declaration of the routines to actually communicate with the server.
// --------------------------------------------------------------------------
Tuple *PL_execute_request(ClientRequest req, int line, const char* file,
			  Tuple * = NULL,
			  long synchrony_type = NOBLOCK,
			  gid &g = (gid&)zero_gid);

void PL_receive_reply(Header &replyHeader, Tuple *&replyTuple) ;


// ----------------------------------------------------------------
// restart flag
// ----------------------------------------------------------------

static long restart_flag = 0;
bool interactive;

// ----------------------------------------------------------------
// fault tolerance mode
// ----------------------------------------------------------------


static Header::FT_ModeType ft_mode = Header::PRIVATE_FT;


static Tuple* private_log = 0;
void PL_proc_start(int, const char*);
void PL_proc_end(int, const char*);


void pl_exit(int val)
{
   PL_proc_end(__LINE__,__FILE__);
   ::exit(val);
}

// ----------------------------------------------------------------
// PLinda main.
// ----------------------------------------------------------------

// user's main.
int real_main(int, char**, char**);

#ifndef alpha
char *rindex(char *s, char c);
#endif
#ifdef linux
char *rindex(char *s, char c) 
{
  if(s == NULL)
    return NULL;

  // find the last occurance of c in the string or NULL
  // algorith is to find the end of s and then work back
  char *p;
  for(p = s ; *p ; p++) 
    ;
  for( ; p >= s && *p != c; p--)
    ;
  if(*p == c)
    return p;
  assert(p < s);
  return NULL;
}
#endif


int need_to_convert;

int main(int argc, char** argv, char** env) {
  if(argc != 6) {
    cerr << "error: wrong arguments, need to run from the plinda server!!\n";
    ::exit(1);
  }
  if(strcmp(argv[4], "interactive")) 
    interactive = true;
  else if (strcmp(argv[4], "noninteractive"))
    interactive = false;
  else {
     cerr << "Wrong 5th arg to main\n";
     exit(1);
  }
  need_to_convert = (int)atoi(argv[5]);
  // the port could also need to be converted, but the daemon or
  // the server had better do this for us!
  assert(need_to_convert == 0 || need_to_convert == 1);

  execname = ::rindex(argv[0], '/');
  if(execname == 0) {
    execname = argv[0];
  } else {
    ++execname;
  }
  ::gethostname(hostname,1000);

  // read the process id.
  // the first argument should contain the process id 
  // that the PLinda server assigns to this process.
  my_pid = procid(atol(argv[1])); 
  my_temp_id = ::getpid();

  // read the address info about the server.
  // The second and third should contain the port number and the ip
  // address of the server whose forms the address. The client
  // will look for the server at the address first and then run a
  // connection protocol.
  unsigned short server_port = (unsigned short)atoi(argv[2]);
  const char* server_host = argv[3];
  server_agent.initialize(my_pid, server_port, server_host,need_to_convert);

  if(server_agent.request_link() == -1) {
#ifdef PL_DEBUG
    cerr << "error: failed to set up a link to the server !\n";
#endif // PL_DEBUG
    ::exit(1);
  }
  PL_proc_start(__LINE__, __FILE__);
  // call the user's main function.
  int val = real_main(argc, argv, env);

  // wait for all the spawned processes to terminate;
  // a client process does not spawn children.
  while(wait(0) != -1);

  // inform the server of termination.
  pl_exit(val);
  assert(0);
  return 0;
}

// --------------------------------------------------------------------
// process anouncements
// --------------------------------------------------------------------

int
PL_num_machines(int line,const char *file)
{
  
  Tuple *tuple = NULL;
  Tuple *reply = PL_execute_request(NUM_MACHINES, line, file, tuple);
  assert(reply && reply->numOfFields() == 1);
  int num;
  reply->read(0,TupleField::PLint, num);
  Tuple::destroy(reply);
  return num;
} 

void PL_stat_begin(int line,char *file)
{
  Tuple *reply = PL_execute_request(STAT_BEGIN, line, file, NULL);
  assert(!reply);
  return;


}

static int theLoop = 0;
static double theSecs = 0;
static Timer myTime;

void PL_pragma_start_trans(int loop)
// this signifies that the same exact transaction will be done loop times
// it will start a timer
// For internal use only!
{
  assert(loop > 0);
  theLoop = loop;
  if(theLoop > 255) {
    cerr << "Implementation shortcoming is can only have 255 iters of loop\n";
    theLoop = 255;
  }
  myTime.clear();
  myTime.start();
  return ;
}
void PL_change_mode(Header::FT_ModeType m, Tuple *oldContinuation,
		    int line, char *file)

{
  ft_mode = m;
  Tuple *reply = PL_execute_request(CHANGE_MODE, line, file, oldContinuation);
  assert(!reply);
  return;

}
void PL_change_interactive(int i,  int line, char *file)
{
  Tuple *send = Tuple::create(1,0);
  assert(send);
  send->setActual(0,TupleField::PLint,i);
  Tuple *reply = PL_execute_request(CHANGE_INTERACTIVE, line, file, send);
  Tuple::destroy(send);
  assert(!reply);
  return;

}

void PL_change_interval(int i,  int line, char *file)
{
  Tuple *send = Tuple::create(1,0);
  assert(send);
  send->setActual(0,TupleField::PLint,i);
  Tuple *reply = PL_execute_request(CHANGE_INTERVAL, line, file, send);
  Tuple::destroy(send);
  assert(!reply);
  return;

}


double PL_pragma_end_trans(int line, char *file)
// this signifies that the xcommit for that transaction is the next statement
// and that C is the size of the continuation
// PL_pragma_trans will send loop, C, Time to the server.
{
  myTime.stop();
  long secs,usecs;
  myTime.total_etime(secs,usecs);
  theSecs = secs + (double)usecs/1000000;
}

void PL_stat_end(int &num,int &interval, int &mode,int &thetime,
		 int line,char *file)
{
  Tuple *reply = PL_execute_request(STAT_END, line, file, NULL);
  assert(reply);
  assert(reply->numOfFields() == 4);
  reply->read(0, TupleField::PLint, num);
  reply->read(1, TupleField::PLint, interval);
  reply->read(2, TupleField::PLint, mode);
  reply->read(3, TupleField::PLint, thetime);
  Tuple::destroy(reply);

  return ;
}


void
PL_proc_eval(Tuple *tuple, int line, const char* file) {

  if(!tuple || !tuple->allActuals()) {
    cerr << "ERROR proc_eval tuple must be present and must be all actuals "<<
      "at line " << line << " in file " << file << endl;
    return;
  }
  tuple->opType(Tuple::EVAL);
  tuple->group(zero_gid);
  Tuple *reply = PL_execute_request(TUPLE_CREATIONS, line, file, tuple);
//  Tuple::destroy(tuple);
  assert(reply == NULL);
} 

bool
PL_arg_rdp(Tuple *tuple, int line, const char* file) {

  Tuple *reply = PL_execute_request(ARG_RDP, line, file,  tuple, NOBLOCK);
  if(reply)  {
    tuple->bindFormals(*reply);
    Tuple::destroy(reply);
    Tuple::destroy(tuple);
    return true;
  }
  Tuple::destroy(tuple);
  return false;
}

void
PL_proc_start(int line, const char* file) {

  Tuple *reply = PL_execute_request(PROC_START, line, file);
  assert(reply == NULL);
}

void
PL_proc_end(int line, const char* file) {

   Tuple *reply = PL_execute_request(PROC_END, line ,file);
   assert(reply == NULL);
}

// --------------------------------------------------------------------
// group manipulation: PL_create_group, PL_destroy_group
// --------------------------------------------------------------------

const gid
PL_create_group(const char *desc, int line, const char* file) {

  assert(desc);
  int len = ::strlen(desc) + 1;
  Tuple *tuple = Tuple::create(1, len * PL_CHAR_SZ);
  tuple->setActual(0,TupleField::PLchar, desc, len);
  gid retValue;
  PL_execute_request(CREATE_GROUP, line, file, tuple, NOBLOCK,retValue);
  // the no block is just put in there so as to be able to use this
  // function, it is bad, but will fix it later
  Tuple::destroy(tuple);
  return retValue;
}

void
PL_destroy_group(gid& g, int line, const char* file) {
#ifndef NDEBUG
  if(g == zero_gid)
    cerr << "PLinda user error trying to destroy an already destroyed group" <<
      " at line " << line << " in file " << file << endl << flush;
#endif
  g = zero_gid;
}

// --------------------------------------------------------------------
// Transactions: PL_xstart, PL_xabort, PL_xcommit
// --------------------------------------------------------------------

void 
PL_xstart(int , const char*) {
  // no fault tolerance is supported.
  if(replayList.length() > 0) {
    return; // we are just in replay mode now!
  }
  if(ft_mode == Header::NO_FT) 
    return; 
  // make sure that nested xstart calls do not take place.
  if(inTransaction) {
    die(E_ALREADY_TRANS);
  }
  needToStartTransaction = true;
  inTransaction = true;
}

void
PL_xcommit(Tuple* log, int line, const char* file) {


#ifndef NDEBUG
  if(log && !log->allActuals()) {
    cerr << "ERROR log tuple must be all actuals at " << line << " in file "
      << file << endl << flush;
  }
#endif

  if(ft_mode == Header::NO_FT)  {
    Tuple::destroy(log);
    private_log = NULL;
    return;
  }
  // make sure the transaction is active
  if(!inTransaction && replayList.length() == 0) {
    die(E_NO_TRANS);
  }
  Tuple *reply = NULL;
  if(log) {
    log->opType(Tuple::XCOMMIT);
    log->group(zero_gid);
  }
  Tuple *sendingLog = NULL;
  if(ft_mode == Header::PRIVATE_FT) {
    Tuple::destroy(private_log);
    // destroy the old log before using the new one
    // since can only keep one around at a time!
    sendingLog = log;
  }

  reply = PL_execute_request(XCOMMIT, line, file, sendingLog);
  // if we got here in REPLAY or GLOBAL mode the old log is not
  // needed since the execute request routine would have give it out if
  // needed.

  Tuple::destroy(private_log);
  private_log = log;
  inTransaction = false;
  assert(reply == NULL);
}

bool
PL_xrecover(Tuple *tuple, int line, const char* file) {
  //  if(replayList.length() > 0 ||  private_log != NULL)  {
  //    cout << "Sorry should only be doing xrecover at start of program\n";
  //    assert(0);
  //  }

  if(!tuple) {
    cerr << "PLinda Error: sorry cannot do xrecover on empty state tuple ";
    cerr << "pattern************\n" << flush;
    pl_exit(1);
  }
  if(ft_mode == Header::NO_FT) 
    return false;
  if(restart_flag == 0) 
    return false; 

  Tuple *reply = PL_execute_request(XRECOVER , line, file, NULL, NOBLOCK);

  if(!reply){
    return false; // no transaction has ever committed
  }
  if(reply->opType() != Tuple::XCOMMIT && reply->opType() != Tuple::FLATTENED){
    // it is just a single tuple of a replay list!

    replayList.append(reply);
    return true;
  }
  if(reply->opType() == Tuple::FLATTENED) {
    unflattenTuples(reply,replayList);
    Pix p;
    p = replayList.first();
    reply = replayList(p);
    if(!reply || reply->opType() != Tuple::XCOMMIT) {
      // there is no state tuple, but there is a replay list
      // return true even though we are not binding anything to this tuple
      // make sure the programmer initializes things before the xrecover!
      return true;
    }
    replayList.del_front();
  }
  assert(reply);
  assert(tuple);

  if(*reply == *tuple)
    tuple->bindFormals(*reply);
  else {
    cerr << "ERROR: xrecovery tuple does not match the pattern of the log tuple\n";
    return false;
  }
  if(ft_mode != Header::PRIVATE_FT) {
    // for global snapshot may need the log tuple
    private_log = reply;
  } else 
    Tuple::destroy(reply);
  Tuple::destroy(tuple);
  return true;
}

// --------------------------------------------------------------------
// Tuple Space: 
//    PL_out
//    PL_in, PL_inp, PL_rd, PL_rdp, 
//    PL_physinp, PL_physrdp, 
// --------------------------------------------------------------------

void
PL_out(gid& g, Tuple* tuple, int line, const char* file) {

 assert(tuple);
#ifndef NDEBUG
 if(!tuple->allActuals()) {
   cerr << "ERROR: trying to out a tuple that is not all actuals on " <<
     line << " in file " << file << endl << flush;
   return;
 }
#endif   
 tuple->opType(Tuple::OUT);
 tuple->group(g);
 Tuple *reply = PL_execute_request(TUPLE_CREATIONS, line, file, tuple, 
				   NOBLOCK,g);
 // The NOBLOCK is totally ignored by everybody, just put it there
 // so don't have to make another execute_request
 // Tuple::destroy(tuple); since out may be buffered do no destroy it here!
 assert(reply == NULL);

}

Tuple *findLocalTuple(Tuple *tuple,bool destroy)
{
  Pix p;
  for(p = theOutList.first(); p ; theOutList.next(p)) {
    Tuple *temp = theOutList(p);
    assert(temp && temp->valid());
    if(*temp == *tuple) {
      // we are inning a tuple we outed in this transaction
      tuple->bindFormals(*temp);
      if(destroy) {
	theOutList.del(p);
	//	Tuple::destroy(temp);
      }
      return temp;
    }
  }
  return NULL;
}
void PL_in(gid& g, Tuple *tuple,int line, const char* file) 
{
  Tuple *temp;
  if((temp = findLocalTuple(tuple,true)) != NULL) {
    if(ft_mode == Header::REPLAY_FT) {
      // must be in recording mode!
      assert(replayList.length() == 0);
      PL_execute_request(RECORD_READ, line, file, temp, BLOCK, g);
      // need to remember this in happened in our replay list
      // do not remember our out that made it happen as this is
      // not needed!
    }	 
    Tuple::destroy(temp);
    return;
  }
  Tuple *reply = PL_execute_request(IN, line, file, tuple, BLOCK, g);
  assert(reply);
  tuple->bindFormals(*reply);
  Tuple::destroy(reply);
  Tuple::destroy(tuple);
}

bool
PL_inp(gid& g, Tuple *tuple,int line, const char* file) {
  Tuple *temp ;
  if((temp = findLocalTuple(tuple,true)) != NULL) {
    if(ft_mode == Header::REPLAY_FT) {
      // must be in recording mode!
      assert(replayList.length() == 0);
      PL_execute_request(RECORD_READ, line, file, temp, BLOCK, g);
      // need to remember this in happened in our replay list
      // do not remember our out that made it happen as this is
      // not needed!
    }	 
    Tuple::destroy(temp);
    return true;
  }
  Tuple *reply = PL_execute_request(IN, line, file, tuple, NOBLOCK,g);

  if(reply) {
    tuple->bindFormals(*reply);
    Tuple::destroy(reply);
    Tuple::destroy(tuple);
    return true;
  }
  Tuple::destroy(tuple);
   return false;
}

void
PL_rd(gid& g, Tuple *tuple, int line, const char* file) {
  if(findLocalTuple(tuple,false)) {
    if(ft_mode == Header::REPLAY_FT) {
      // must be in recording mode!
      assert(replayList.length() == 0);
      PL_execute_request(RECORD_READ, line, file, tuple, BLOCK, g);
      // need to remember this in happened in our replay list
      // do not remember our out that made it happen as this is
      // not needed!
    }	 
    return ;
  }
   Tuple *reply = PL_execute_request(RD, line, file, tuple,BLOCK,g);
   assert(reply);
   tuple->bindFormals(*reply);
   Tuple::destroy(reply);
   Tuple::destroy(tuple);
}

bool
PL_rdp(gid& g, Tuple *tuple, int line,  const char* file) {
  if(findLocalTuple(tuple,false)) {
    if(ft_mode == Header::REPLAY_FT) {
      // must be in recording mode!
      assert(replayList.length() == 0);
      PL_execute_request(RECORD_READ, line, file, tuple, BLOCK, g);
      // need to remember this in happened in our replay list
      // do not remember our out that made it happen as this is
      // not needed!
    }	 
    return true;
  }
  Tuple *reply = PL_execute_request(RD, line, file, tuple,NOBLOCK,g);
   if(reply) {
     tuple->bindFormals(*reply);
     Tuple::destroy(reply);
     Tuple::destroy(tuple);
     return true;
  }
   Tuple::destroy(tuple);
   return false;
}


// ---------------------------------------------------------------------
// Utility routines 
// ---------------------------------------------------------------------


Tuple *handleReplay(ClientRequest req, Tuple *tuple, long synchrony_type,
		    gid &g)
{
     Pix p;
     Tuple *reply = NULL;
     p  = replayList.first();
     // get the head of the list now, but may not actually use it
     switch(req) {
     case TUPLE_CREATIONS:
       assert(tuple); 
       Tuple::destroy(tuple);
       break;
     case IN:
     case RD:
     case ARG_RDP:
       reply = replayList(p);
       replayList.del_front();
       assert(synchrony_type == NOBLOCK || reply);
       break;
     case NUM_MACHINES:
       reply = replayList(p);
       replayList.del_front();
       assert(reply && reply->numOfFields() == 1);
       break;
     case CREATE_GROUP:
       reply = replayList(p);
       replayList.del_front();
       assert(reply&&reply->numOfFields()==0&&reply->length()==sizeof(Tuple));
       assert(reply->opType() == Tuple::GROUP);
       g = reply->group();
       Tuple::destroy(reply);
       break;
     case XCOMMIT:
       reply = replayList(p);
       replayList.del_front();
       assert(tuple == NULL);
       assert(reply == NULL);
       break;
     case CHANGE_MODE:
       cerr << "replaying change_mode does nothing!!\n" << flush;
       break;
     case CHANGE_INTERVAL:
       cerr << "replaying change_interval does nothing!!\n" << flush;
       break;
     case CHANGE_INTERACTIVE:
       cerr << "replaying change_interactive does nothing!!\n" << flush;
       break;
     default:
       cout << "unexpected req in client ****** " << req << endl << endl ;
       assert(0);
     }
     if(replayList.length() == 0) {
       assert(req == XCOMMIT);
     }
     return reply;
}

Tuple *PL_execute_request(ClientRequest req, 
			  int line, const char* file,Tuple *tuple,
			  long synchrony_type,
			  gid& g) 
			 
{		      

  Tuple *replyTuple = NULL;
  Header replyHeader;
  Header header;
  if(replayList.length() > 0) {
     return handleReplay(req, tuple, synchrony_type, g);
   }
#ifndef NDEBUG
   if(req!= PROC_START && req != PROC_END &&
      req != XRECOVER  && req != XCOMMIT && req != NUM_MACHINES &&
      req != STAT_BEGIN && req != STAT_END && req != CHANGE_MODE) {
     assert(tuple);
     assert(tuple->valid());
   } else if(req != XCOMMIT)
     assert(tuple == NULL);
#endif
   if(req == TUPLE_CREATIONS && inTransaction && 
      tuple && tuple->length() < LARGE_TUPLE) {
     assert(tuple);
     assert((tuple->opType() == Tuple::EVAL  && tuple->group() == zero_gid)||
	    (tuple->opType() == Tuple::OUT  && tuple->group() != zero_gid));
     theOutList.append(tuple);
     return NULL;
   } 
   header.clientRequest(req);
   header.transientId(my_temp_id);
   header.fileName((char*)file);
   header.lineNum(line);
   header.synchronyType(synchrony_type);
   header.g(g);
   
   if(req == XCOMMIT) {
     if(tuple)
       theOutList.append(tuple);
     tuple =  flattenTuples(theOutList,need_to_convert,0);
     // definitely want to delete tuple after this!
     assert(theOutList.length() == 0);
     if(theLoop) {
       // this is a profiled transaction, set it up
       header.secs(theSecs);
       header.loop(theLoop);
     } else
       header.loop(0); // indicate not a profiling transaction
   }
   const int tupleLength = tuple ? tuple->length() : 0;
   header.tupleLength(tupleLength);


   header.xstart(needToStartTransaction);
   needToStartTransaction = 0;
   assert(true != 0);
   assert(false == 0);
   if(req == CHANGE_MODE) 
     header.ftMode(ft_mode);

   if(server_agent.send(header,tuple) == -1) {
      die(E_SEND_FAIL);
   }
   if(req == STAT_BEGIN ||  req == CHANGE_INTERACTIVE 
      || req == CHANGE_INTERVAL || req == CHANGE_MODE)
     return NULL;
   PL_receive_reply(replyHeader, replyTuple);
   assert((replyHeader.tupleLength() == 0 && replyTuple == NULL) ||
	  replyTuple && replyHeader.tupleLength() == replyTuple->length());
   if(req == PROC_START) {
      // fill up some local stuff, I know this is bad to do it here,
      // but it saves returning a useless Header to all the functions
      // except PROC_START
      restart_flag = replyHeader.restartFlag();
      ft_mode = replyHeader.ftMode();
      assert(replyTuple == NULL);
   } else if(req == CREATE_GROUP) {
      g = replyHeader.g();
   }  else if(req == XCOMMIT) {
     ft_mode = replyHeader.ftMode();
   } else if((req == TUPLE_CREATIONS || req == XCOMMIT) && tuple &&
	     tuple->opType() != Tuple::XCOMMIT) {
     Tuple::destroy(tuple);
   }

   return replyTuple;
}

void PL_receive_reply(Header &replyHeader, Tuple *&replyTuple)

{
   while(1) {
     replyHeader.reset();
     replyTuple = NULL;
     // receive a reply
      if(server_agent.receive(replyHeader,replyTuple) == -1) {
	 die(E_RECEIVE_FAIL);
      }
#ifndef NDEBUG
      if(!replyHeader.status()) {
	 cerr << "In the client got bad news from the server " <<
	   replyHeader.status().message() << endl << flush;
	 die(replyHeader.status());
      }
#endif
      assert(replyHeader.type() == Header::SEND_CLIENT_MESSAGE);
      if(replyHeader.clientRequest() == LOG_REQ) {
	// startTimer Hansoo
	assert(replyTuple == 0 && replyHeader.tupleLength() == 0);
	int logLength = private_log ? private_log->length() : 0;
	if(private_log) {
	  assert(private_log->opType() == Tuple::XCOMMIT);
	}
	 Header header;
	 header.clientRequest(LOG_FLUSH);
	 header.transientId(my_temp_id);
	 header.tupleLength(logLength);
	 header.xstart(0);
	 if(server_agent.send(header, private_log) == -1) {
	    die(E_SEND_FAIL);
	 }
	 // end timer Hansoo
      } else 
	 break;
   } // while
}
