/*********
File:			Emit_PLinda.c
Purpose:		This file emits the proper PLinda code so it 
			can be linked and compiled with the existing
			PLinda source code written in C.
Last Altered:		11/20/96
Author:			Tom Brown(brow7141@cs.nyu.edu)
*********/
#include <stdio.h>
#include <assert.h>
#include "SymTable.h"
#include "Match_keywords.h"
#include "List.h"
#include "Emit_PLinda.h"
#include "plinda_f.h"

#define NO_FURTHER_EMITTANCE_NEEDED 5
#define GID_EXPECTED 1
#define GID_NOT_EXPECTED 0

extern char * File_Name;



/******
Function:	Emit_Line_Number_and_File_Name
Purpose:	Emit target code for a PLinda function call with the line 
		number and the source file name.
******/
void
Emit_Line_Number_and_File_Name(CONTEXT * c, FILE * plinda)
{
	Out_PLinda_Int(c->line_number, plinda);	
	Out_PLinda(",", plinda);	

	Out_PLinda("\"", plinda);	
	Out_PLinda(File_Name, plinda);	
	Out_PLinda("\"", plinda);	

	Out_PLinda(")\n",plinda);	
}



/******
Function:	Emit_xstart
Purpose:	Emit target code for PLinda command xstart().
******/
void
Emit_xstart(CONTEXT * c, FILE * plinda)
{
	Out_PLinda("        call fxstart(",plinda);
	Emit_Line_Number_and_File_Name(c, plinda);
}



/******
Function:	Emit_xcommit_stateless
Purpose:	Emit target code for PLinda command xcommit() with no tuple space
		arguments specified.
******/
void
Emit_xcommit_stateless(CONTEXT * c, FILE * plinda)
{
	Out_PLinda("       call fxcommit_nostate(", plinda);
	Emit_Line_Number_and_File_Name(c, plinda);
}



/******
Function:	Emit_destroy_group
Purpose:	Emit target code for PLinda command destroy_group().
******/
void
Emit_destroy_group(CONTEXT * c, FILE * plinda)
{
	char temp;
	Out_PLinda("        call fdestroy_group(",plinda);

	CON_Align_to_Char('(', c);

	temp = CON_getc(c);
	while (temp != ')')
	{
		Out_PLinda_char(temp, plinda);	
		temp = CON_getc(c);
	}

	Out_PLinda(",", plinda);	
	Emit_Line_Number_and_File_Name(c, plinda);
}



/******
Function:	Emit_create_group
Purpose:	Emit target code for PLinda command destroy_group().
Called From: 	Emit_Stateless_PLinda_Function
******/
void
Emit_create_group(CONTEXT * c, FILE * plinda)
{
	char temp;
	Out_PLinda("        call fcreate_group(",plinda);

/** 	CON_print(c); **/

	CON_Align_to_Char('(', c);

	temp = CON_getc(c);
	while (temp != ')')
	{
		Out_PLinda_char(temp, plinda);	
		temp = CON_getc(c);
	}

	Out_PLinda(",", plinda);	
	Emit_Line_Number_and_File_Name(c, plinda);
}





/******
Function:	Emit_Stateless_PLinda_Function
Purpose:	Emit PLinda function call according to parsing mode in its stateless format.
		Currently, the only stateless function is xcommit(), so if not xcommit,
		then an error has occurred -Output error to <stdout>!!
Called From:	Emit_Pl_Command() of <Analyze_for_Emittance.c>
******/
void
Emit_Stateless_PLinda_Function(int parsing_mode, CONTEXT * c, FILE * plinda)
{
/**
	fprintf(stderr,"Emit_Stateless:: parsing_mode => %d\n", parsing_mode);
**/
        switch (parsing_mode)
        {
                case XSTART:
                        Emit_xstart(c, plinda);
                        break;
                case DESTROY_GROUP:
                        Emit_destroy_group(c, plinda);
                        break;
                case CREATE_GROUP:
                        Emit_create_group(c, plinda);
                        break;
                case XCOMMIT:
			Emit_xcommit_stateless(c, plinda);
                        break;
		default:
			fprintf(stderr,"Error:: Line %d ", c->line_number);
			fprintf(stderr,"Unexpected call in Emit_Stateless_\n");
			CON_print(c);
			fprintf(stderr,"\n");
	}
}



/******
Function:	Emit_Sundry_PLinda_calls
Purpose:	Emit various PLinda function calls
Called From:	Emit_Pl_Command() of <Analyze_for_Emittance.c>
******/
void
Emit_Sundry_PLinda_calls(int parsing_mode, CONTEXT * c,FILE * plinda)
{
	switch (parsing_mode)
	{
		case XCOMMIT:
			Out_PLinda("       call fxcommit(",plinda);
			break;
                case DESTROY_GROUP:
			Out_PLinda("       call fdestroy_group(",plinda);
                        break;
                case CREATE_GROUP:
			Out_PLinda("       call fcreate_group(",plinda);
                        break;
		case XRECOVER:
			Out_PLinda("       call fxrecover(",plinda);
			break;
		case PROC_EVAL:
			Out_PLinda("       call fproc_eval(",plinda);
			break;
		case ARG_RDP:
			Out_PLinda("       fargrdp(",plinda);
			break;
		case PL_OUT:
			Out_PLinda("       call fout(",plinda);
			break;
		case PL_IN:
			Out_PLinda("       call fin(",plinda);
			break;
		case PL_INP:
			Out_PLinda("       call finp(",plinda);
			break;
		case PL_RD:
			Out_PLinda("       call frd(",plinda);
			break;
		case PL_RDP:
			Out_PLinda("       call frdp(",plinda);
			break;
                default:
                        fprintf(stderr,"Emit_Sundry_PLinda:: Error Line => ");
                        CON_print(c);
	}
}




/*****
Function:	Emit_Beginning_of_Tuple_Space_Info
Purpose:	Emits the typical information (number of tuple arguments) to
		begin a PLinda tuple space function call.  Note, the line
		continuation emitted at the bottom, '+' symbol, is done for
		aesthetic value for reading the output file.
Called From:	Emit_Pl_Command() of <Analyze_for_Emittance.c>
*****/
void
Emit_Beginning_of_Tuple_Space_Info(int tuple_length, FILE * plinda)
{
	Out_PLinda("ftuple(",plinda);	
	Out_PLinda_Int(tuple_length,plinda);
	Out_PLinda(",\n",plinda);	
	Out_PLinda("     +  ",plinda);	
}


/*****
Function:	Emit_End_of_Tuple_Space_Info
Purpose:	Emits the typical information (line number, File_Name) to
		end a PLinda tuple space function call.
Called From:	Emit_Pl_Command() of <Analyze_for_Emittance.c>
*****/
void
Emit_End_of_Tuple_Space_Info(CONTEXT * c, FILE * plinda)
{
	Out_PLinda("),\n",plinda);	
	Out_PLinda("     +  ",plinda);	
	Out_PLinda_Int(c->line_number, plinda);	
	Out_PLinda(",", plinda);	

	Out_PLinda("\"", plinda);	
	Out_PLinda(File_Name, plinda);	
	Out_PLinda("\"", plinda);	

	Out_PLinda(")\n",plinda);	
}



/*******
Function:       Emit_According_to_Var_Type
Purpose:        Emits the tuple-space type code specific to the 'var_type'
                of the variable passed in.
Return:         Returns 1 if the no more output needed in the Tuple Space
                area, 0 if other output like the variable name and array size
		still needs to be outputted in the "Analyze_".
Called From:    Analyze_for_Type_and_Size_Emittance()
*******/
int
Emit_According_to_Var_Type(int var_type, char var_name[80], int l_value, 
						FILE * plinda, CONTEXT * c)
{
/**
	fprintf(stderr,"\nEmit_According_to_Var_Type:: var_name => %s", var_name);
	fprintf(stderr,"; var_type => %d\n\n", var_type);
**/
        switch (var_type)
        {
                case GID:
                        Emit_GID(var_name, l_value, plinda);
                        return 1;
                case INTEGER:
                        if (l_value == 1)
                                Out_PLinda("flint(",plinda);
                        else
                                Out_PLinda("frint(",plinda);
                        break;
                case REAL:
                        if (l_value == 1)
                                Out_PLinda("flfloat(",plinda);
                        else
                                Out_PLinda("frfloat(",plinda);
                        break;
                case DOUBLE:
                        if (l_value == 1)
                                Out_PLinda("fldouble(",plinda);
                        else
                                Out_PLinda("frdouble(",plinda);
                        break;
                case CHARACTER:
                        if (l_value == 1)
                                Out_PLinda("flchar(",plinda);
                        else
                                Out_PLinda("frchar(",plinda);
                        break;
                default:
                        fprintf(stderr,"Emit_Stateless_PLinda:: Error Line ");
                        fprintf(stderr,"%d: '%s' - Variable of Bad Type.\n",
                                                c->line_number, var_name);
                        fprintf(stderr,"Emit_Stateless_PLinda:: Error Line => ");
                        CON_print(c);
        		return 1;
        }
        return 0;
}



/******
Function:       Emit_GID
Purpose:        Emit code for the special gid type into the tuple space
                parameters for a PLinda function call
                (eg. pl_out(<gid_code>, ...); )
Called From:    Emit_According_to_Var_Type()
******/
void
Emit_GID(char var_name[80], int l_value, FILE * plinda)
{
        if (l_value == 1)
                Out_PLinda("flint(",plinda);
        else
                Out_PLinda("frint(",plinda);
        Out_PLinda(var_name, plinda);
        Out_PLinda(",2)",plinda);
}



/******
Function:       Emit_Constant_String
Called From:	Constants_Emit()
******/
void Emit_Constant_String(char temp[80], FILE * plinda)
{
	Out_PLinda("frstr(",plinda);
	Out_PLinda(temp,plinda);
	Out_PLinda(")",plinda);
}



/******
Function:       Emit_Constant_Int
Called From:	Constants_Emit()
******/
void Emit_Constant_Int(char temp[80], FILE * plinda)
{
	Out_PLinda("frint(",plinda);
	Out_PLinda(temp,plinda);
	Out_PLinda(",0)",plinda);
}


/******
Function:       Emit_Constant_Float
Called From:	Constants_Emit()
******/
void Emit_Constant_Float(char temp[80], FILE * plinda)
{
	Out_PLinda("frfloat(",plinda);
	Out_PLinda(temp,plinda);
	Out_PLinda(",0)",plinda);
}



/******
Function:       Emit_Constant_Double
Called From:	Constants_Emit()
******/
void Emit_Constant_Double(char temp[80], FILE * plinda)
{
	Out_PLinda("frdouble(",plinda);
	Out_PLinda(temp,plinda);
	Out_PLinda(",0)",plinda);
}


