/********
File:		Match_keywords.c 
Last Revision:	5/5/96
Author:		Tom Brown (brow7141@cs.nyu.edu)
Purpose:	This file is the main driver of the lex/parsing
		engine of the f77/plinda compiler.
Theme: 		The overall theme of this file and the program
		is to collect information about var. declarations
		to be used in the PLinda output code.
		Therefore, all Fortran code is echoed out to file
		<plinda.f>.  The PLinda code is created in the file
		<Emit_PLinda.c>.  Recognition of the PLinda keywords
		takes place in function lex_key_word() at the bottom
		of the program.
********/
#include "Match_keywords.h"
#include "Match_Tools.h"
#include "Char_Match.h"
#include "ParsingRules.h"
#include "DebugError.h"
#include "SymTable.h"
#include "List.h"
#include "plinda_f.h"
#include "Analyze_for_Emittance.h"
#include "y.tab.h"		/** For token constants being returned...**/

#define ASCII_TO_INTEGER 48	/*  To Convert Ascii digits to actual digits */
#define DECORATE 1		/*  To Comment out PLinda keywords	*/
#define GENERATE 0		/*  Actively translate Fortran/PLinda  	  */
#define MAX_LEN_KEY 15		/*  Maximum length of all the keywords	  */

extern int line_number;
static int array_collect_mode = 0;	/* Count number of array dimensions*/ 
static SymTable * SymbolTable = NULL;	




/******
Function:	TerminateVarStr
Called From:	Collection()
Purpose:	Returns 1 if the character passed is a variable terminator.
		There are two characters that can terminate a variable in
		Fortran: '\n' and ','
		End of Line terminates the variable if the next line
		does not contain a line continuation symbol.
Note:		This function will also count the number of array dimensions
		using array_collect_mode a counter.
******/
int
TerminateVarStr(CONTEXT * c, char pass, int parsing_mode, FILE * plinda)
{
	switch (pass)
	{
		case ',':	
			if ((parsing_mode == SUBROUTINE) ||
					(array_collect_mode > 0))
				return 0;
			else
				return 1;

		case '\n': 	if (Line_Continuation(c) == 0)
				{	
/**
					Out_PLinda(c->in_buff, plinda);
					CON_next_line(c);
**/
					return 1;
				}
				else
				{
					Out_PLinda(c->in_buff, plinda);
					CON_next_line(c);
					return 0;
				}
		default:	return 0;
	}
}





/******
Function:	PossibleVariableChars
Called From:	Collection()
Purpose:	Returns 1 if the character passed in is a legal character 
		in a variable string.  For instance, a comma is used to
		separate variable declarations from one another, and thus
		is not a legal character in a variable string, so this program
		will just spit/echo it back out and continue on as if nothing
		is wrong.
Note:		array_collect_mode is a global variable collecting info.
		about the number of dimensions of the array. 
******/
int
PossibleVariableChars(char pass, int parsing_mode)
{
	if ((Alpha(pass)) || (Number(pass)))
		return 1;
	switch (pass)
	{
		case '-':
		case '+':
		case '/':
		case ':':	
				if (array_collect_mode > 0)
				{
					return 1;
				}
				else
				{
					return 0;
				}
		case ',':	if (array_collect_mode > 0)
				{
					array_collect_mode++;
					return 1;
				}
				if (parsing_mode == SUBROUTINE)
					return 1;
				return 0;	/*** Means new variable!!**/
		case '(':	
				if (parsing_mode != SUBROUTINE)
				{
					array_collect_mode = 2;
					return 1;
				}
				else
					return 0;
		case ')':  	/** The ')' might mean the end of an array **/
				array_collect_mode--;
				return 0;
		case '_':	
				return 1;
		default: 	
				return 0;
	}
}




/*******
Function:	Collection_Subroutine
Purpose:	Called from lex_key_word upon recognizing a keyword "SUBROUTINE"
		This function basically collects the subroutine name, and its 
		parameters (even though in Fortran we don't care about them
		because the parameter vars. are re-declared.
*******/
void
Collection_Subroutine(int parsing_mode, CONTEXT * c, FILE * plinda)
{
	int i, j;
	char temp[80];

	for (i = 0; i < 80; i++)
  		temp[i] = '\0';

	i = 0;
	temp[i] = CON_getc(c);
	while (temp[i] != '(')
	{    /** Collect a token -Parse until seeing a token or something) **/
		if (!(White_Space(temp[i])))
			i++;
		temp[i] = CON_getc(c);
		if (End_of_Line(temp[i]))
		{
 			if (Line_Continuation(c) == 0)
				break;
			Out_PLinda(c->in_buff, plinda);
			CON_next_line(c);
		}
	}
	/** Store the subroutine name.	**/
	temp[i] = '\0';				/* Remove the extra char.*/
	Store_token(temp, parsing_mode, 0, 0);
	for (j = 0; j < i; j++)
  		temp[j] = '\0';
	i = 0;
	temp[i] = CON_getc(c);

	/** We don't need to parse the parameters of the subroutine in
		since in Fortran77 the types of the parameters are 
		immediately declared below the Subroutine header.
	**/
}



/*****
Function:	Collection_COMMON
Purpose:	Collect the token information about the Common block
*****/
void
Collection_COMMON(CONTEXT * c, FILE * plinda)
{
	char test_char;

	test_char = CON_getc(c);
	if (test_char == '/')
	{
	/* Gather the Label of the Common Block (for the trash - not used) */ 
		test_char = CON_getc(c);
		while (test_char != '/')
			test_char = CON_getc(c);
	}

/* Now, gather each token according to the rule => defined in Collection*/ 	
	Collection(COMMON, c, plinda);
}



/*****
Function:	Collection_IMPLICIT
Purpose:	To implement the *crazy* IMPLICIT keyword.
Note:		There are a few different ways to declare the IMPLICIT
		stmt.  Below are the ways to do it. 
Idea:	 	Regular Expr :
		implicit <type> (letterI-letterJ (,letterI2-letterJ2)*)
		implicit <type> (letterI (,letterJ)* )
		implicit <type> (letterI (,letterJ)* ), <type> (letterI-letterJ)
*****/
void
Collection_IMPLICIT(CONTEXT * c, FILE * plinda)
{
	SymTable * presentSymTable;
	char check;
	int get_type, size;

/** 1. Get the implicit type decl.	**/
	get_type = Get_Type(c);
	if (get_type == 0)		/** Implicit NONE **/
	{
		presentSymTable = Present_SymTable(SymbolTable);
		Insert_List_Implicit(presentSymTable->implicit, 0, 0, 0, 0); 
		return;
	}

/** 2. Optional Length Specification 	**/
	size = Opt_Length_Spec(c, plinda);
	if (size != 0)
		get_type = Adjust_Type_By_Size(get_type,size);


/* 3. Now Collect the letter range specifications in the (...) .
	And, add letter range to a entry in the symTable implicit list.*/
	Collect_Letter_Range(c, get_type, plinda);

	while (!(End_of_Line(check)))
	{
		check = CON_getc(c);
	}
}



/******
Function:	Get_Type
Purpose: 	This function reads the next token in the context buffer
		and returns what type it is - we are guaranteeing that
		the type will be of type: (int,real,etc.)
Called From: 	Collection_IMPLICIT()
******/
int
Get_Type(CONTEXT * c)
{
	char build[MAX_LEN_KEY];
	int i = 0;
	int test = 0;
	build[i] = CON_getc(c);
	while (test == 0)
	{
		if (White_Space(build[i]) == 0)
			i++;	
		build[i] = CON_getc(c);
		if (End_of_Line(build[i]))
			return (0);
		if (i > 2)		/* No keyword <= 3 chars.*/
			test = MatchType(build); 
	}	
	if (test == -1)
		return (0);		/* Implicit NONE */
/**
fprintf(stderr,"Type of the impliciti is => %d\n",test);
**/
	return (test);
}



/******
Function:	MatchType
Purpose: 	This function reads the next token in the context buffer
		and returns what type it is - we are guaranteeing that
		the type will be of type: (int,real,etc.)
Note:		If an 'implicit	none' is declared then we know no variables
		to be implicit variables in the program.
******/
int
MatchType(char str[MAX_LEN_KEY])
{
	if (keyword_test("none", str, 4))
		return (-1);
	if (keyword_test("real", str, 4))
		return (REAL);
	if (keyword_test("integer", str, 7))
		return (INTEGER);
	if (keyword_test("character", str, 9))
		return (CHARACTER);
	if (keyword_test("complex", str, 7))
		return (COMPLEX);
	if (keyword_test("doubleprecision", str, 15))
		return (DOUBLE);
	if (keyword_test("logical", str, 7))
		return (LOGICAL);
	return (0);
}




/******
Function:	Opt_Length_Spec
Purpose: 	Optional Length Specifications in Fortran of the form:
				variable_decl * integer 
		(ie. real*8 or character * 123)
Returns:	This function returns the length specified.

Delicate:	This function doesn't report any errors if it sees/parses
		anything very strange.  This strategy is done based on the
		fact that this function may be called from a lot of 
		different places.
******/
int
Opt_Length_Spec(CONTEXT * c, FILE * plinda)
{
	char test; 
	int length = 0;

	test = Collect_Single_Letter(c);
	if (test != '*')
	{
/**
	fprintf(stderr,
	  "Error: Opt_Length_Spec:: Expected '*' instead of '%c'!\n",test);
**/
		CON_backup(c);
		return (0);
	}
	test = Collect_Single_Letter(c);
        while (White_Space(test) != 0)
	{
		test = Collect_Single_Letter(c);
	}
	while (Number(test) != 0)
	{
		length = (length * 10) + Get_Number(test);
		test = CON_getc(c);
        	while (White_Space(test))
			test = Collect_Single_Letter(c);
	}
	CON_backup(c);
	return (length);
}



/*****
Function: 	Adjust_Type_By_Size
Purpose:	The optional length specification given - now, adjust
		the variable by size.
		Return the new type the variable should be according
		to its size. 
Default:	Just ignore the size and return the type specified.
*****/
int
Adjust_Type_By_Size(int type, int size)
{
	if ((type == REAL) && (size == 8))
		return REAL;
	if ((type == REAL) && (size == 16))
		return DOUBLE;
	return (type);
}



/*******
Function:	Collect_Letter_Range
Purpose:	Collect the letter range specification of the IMPLICIT 
		statement.
		(ie. 'implicit real(a-f)')
		(ie. implicit <type> (letterI-letterJ (,letterI2-letterJ2)*)
		(ie. implicit <type> (letterI (,letterJ)* )
*******/
void
Collect_Letter_Range(CONTEXT * c, int type, FILE * plinda)
{
	SymTable * presentSymTab;
	char test, range_a, range_b;
	int implicit_type;

/* 1. Collect the range info. */ 
	test = Collect_Single_Letter(c);
/***
	if (test != '(')
		fprintf(stderr," Collect_Lett=> %c, not left paren\n",test);
***/

	range_a = Collect_Single_Letter(c);

	test = Collect_Single_Letter(c);
	if (test == ',')
	{
		range_b = Collect_Single_Letter(c);
		implicit_type = 0;
	}
	if (test == '-')
	{
		range_b = Collect_Single_Letter(c);
		implicit_type = 1;
	}

/* 2. Store the range info. */ 
	presentSymTab = Present_SymTable(SymbolTable);
	presentSymTab->implicit = Insert_List_Implicit(presentSymTab->implicit,
					range_a, range_b, type, implicit_type);
	test = Collect_Single_Letter(c);

/* 3. Case: implicit real(a,b,c,d)  - Collect all the letters */
	while (test != ')')
	{
		if (test == ',')
		{
			range_a = Collect_Single_Letter(c);
			test = range_a;
		}
		if (test == ')')
		{
			range_b = 0;
		}
		else
		{
			test = Collect_Single_Letter(c);
			if ((test == ',') || (test == '-'))
			{
				range_b = Collect_Single_Letter(c);
			}
			test = Collect_Single_Letter(c);
		}
		presentSymTab = Present_SymTable(SymbolTable);
		presentSymTab->implicit = Insert_List_Implicit(presentSymTab->implicit,
					range_a, range_b, type, implicit_type);
	}

/* 4. Look for any other implicit stmts. made on that line */
	test = Collect_Single_Letter(c);
	if (End_of_Line(test))
	{
		if (Line_Continuation(c) == 0)
			return;
		Out_PLinda(c->in_buff, plinda);
		CON_next_line(c);
	}
/* 5. Recursive!! Start the implicit collecting process again. */
		Collection_IMPLICIT(c, plinda);
}




/*******
Function:	Collection
Purpose:	Called from lex_key_word upon recognizing a keyword
		that we care about (ie. integer, pl_out, etc.)
		This function basically collects the tokens thru the
		end of the line, taking the proper action according
		to the 'parsing_mode' variable.
Theme:		Grammar stmt of this parsing should look something like
		one of two things:
		(1)Token : 	variable_string  <, Token>
		(2)Token :      :: variable_string = const_literal, <Token>
		Corresponding Regular Expression:
				variable_string (, variable_string)*
Applied to:	This function is applied to Integers, Reals, Double Precision,
		And, it takes into account that there may be arrray decls. in
		here too, so to ignore whatever vars./consts. that are inside the ().
		This is taken into account by PossibleVariableChars().
Last Alteration: 	1/29/96
********/
void
Collection(int parsing_mode, CONTEXT * c, FILE * plinda)
{
	int i, j, length_spec, range_a, range_b, line_cont;
	char temp[80], check;
	for (i = 0; i < 80; i++)
  		temp[i] = '\0';
	range_a = 0;
	range_b = 0;
	line_cont = 0; 
	i = 0;
#ifdef PARSE_DEBUG
 fprintf(stderr,"Collection:: line parsed => %s\n",c->in_buff); 
#endif

	temp[i] = CON_getc(c);
	while (1)
	{
		length_spec = 0;
#ifdef PARSE_DEBUG
 fprintf(stderr,"Collection:: TerminateVarStr on char => %c\n",temp[i]); 
#endif
	
		while (!(TerminateVarStr(c, temp[i], parsing_mode, plinda)))
		{
			if (PossibleVariableChars(temp[i], parsing_mode))
			{
			 	if ((array_collect_mode==0) && 
					(!((i==0)&&(Number(temp[i])))))
						i++;
				else
				{ /*  Collect Array Size */
					if (temp[i] == ':')
					{
						range_a = range_b;
						range_b = 0;
					}
					else
					{
					   if (temp[i] != '(') 
						range_b = (range_b * 10) + 
								(temp[i] - 48);
					}	
				}
			}
			else
			{
				if (temp[i] == '*')
				{
					CON_backup(c);
					length_spec = Opt_Length_Spec(c,plinda);
				}
				if (temp[i] == '=')
				{
					/** Assignment at Declaration Time **/
					while (!TerminateVarStr(c, temp[i], 
							  parsing_mode, plinda))
						temp[i] = CON_getc(c);
					temp[i] = '\0';
					break;
				}
			}
			temp[i] = CON_getc(c);
		}
#ifdef PARSE_DEBUG
fprintf(stderr,"Collection:: collected_variable_string=> %s\n",temp);
#endif
		check = temp[i];
		if (i > 0)
			temp[i] = '\0';
#ifdef PARSE_DEBUG
 	fprintf(stderr,"Collection =>lengthspec-> %d\n",length_spec);
	fprintf(stderr,"Collection;; range_a => %d ; range_b => %d\n",
							range_a, range_b);
#endif

		Store_token(temp,parsing_mode,array_collect_mode,length_spec);
		array_collect_mode = 0;	
		if (End_of_Line(check))
		{
			if (Line_Continuation(c) == 0)
			{
				break;
			}
			Out_PLinda(c->in_buff, plinda);
			CON_next_line(c);
		}
		for (j = 0; j <= i; j++)
  			temp[j] = '\0';
		i = 0;
	}
/** Print_SymTable(SymbolTable); **/
}



/*******
Function:	Store_token
Purpose:	Store the token according to its type - if a subroutine, then
		store it in the name field of the symbol_table ; if a variable
		(ie. int) store it in the list off of the symbol table.
Note:
Not always true that the last block is the present block -
Think about records - Find upon the record/subroutine name
*******/
void
Store_token(char * token, int type, int arr_size, int opt_len)
{
	SymTable * presentSymTable;
	if (type == SUBROUTINE) 
		presentSymTable = Last_SymTable(SymbolTable);
	if (type != SUBROUTINE)	/* first subroutine creates the SymbolTable */
	{
		presentSymTable = Present_SymTable(SymbolTable);
		if (Error((SymTable *)presentSymTable,"Store_token","presentSymTable"))
			return;
	}	
	switch (type)
	{
		case INTEGER:	
		case REAL:	
		case CHARACTER:	
		case DOUBLE:	
		case COMMON:	
		case GID:	
		  presentSymTable->list = Insert_List(presentSymTable->list, 
						  token, type, arr_size, opt_len);
			break;	
		case SUBROUTINE:	
			SymbolTable = Insert_SymTable(presentSymTable, token);
			break;
	}
}



/*****
Function:	Collection_Include
Purpose:	This function is called upon the keyword 'include' being 
		recognized.  If after the include is parsed, a filename 
		is not listed in quotes then an error is assumed and the 
		function returns quietly (no error messages).
*****/
void
Collection_Include(CONTEXT * c, FILE * plinda)
{
	FILE * include_file;
	CONTEXT * new_context_buffer;
	char temp[100];
	int i = 0;

	temp[i] = CON_getc(c);
	while (!((temp[i] == '\'') || (temp[i] == '\"')))
	{
		if (End_of_Line(temp[i]))
		{
			if (Line_Continuation(c) == 0)
				return;
			Out_PLinda(c->in_buff, plinda);
			CON_next_line(c);
		}
		temp[i] = CON_getc(c);
	}

	temp[i] = CON_getc(c);
	temp[i+1] = '\0';
	while (!((temp[i] == '\'') || (temp[i] == '\"')))
	{
		if (End_of_Line(temp[i]))
		{
			if (Line_Continuation(c) == 0)
				break;
			Out_PLinda(c->in_buff, plinda);
			CON_next_line(c);
		}
		i++;
		temp[i] = CON_getc(c);
		temp[i+1] = '\0';
	}
	temp[i] = '\0';

	if ((include_file = fopen(temp,"r")) == NULL)
	{
		fprintf(stderr,"Error: Couldn't open file %s\n",temp);
		return;
	}
	
	new_context_buffer = CON_create(include_file);
	Out_PLinda("\n",plinda);
	Out_PLinda("C       include \"",plinda);
	Out_PLinda(temp, plinda);
	Out_PLinda("\"\n",plinda);
	while (1)
	{
		lex_key_word(new_context_buffer, plinda, GENERATE);
		if (!CON_next_line(new_context_buffer))
			break;
		else
			line_number++;
	}
	Out_PLinda("\n",plinda);
	Out_PLinda("C       end \"",plinda);
	Out_PLinda(temp, plinda);
	Out_PLinda("\"\n",plinda);
}




/*****
Function:	ParseKeywords
Purpose:	Parsing action taken based upon the keyword parsed in lex_key_word
		This function is called with both PLinda keywords and Fortran variable
		declarations.
Called From:	lex_key_word()
Last Altered:	1.27.96
*****/
void
ParseKeywords(int parsing_mode, CONTEXT * c, FILE * plinda)
{
	Line_Number(c, plinda);

	switch (parsing_mode)
	{
		case SUBROUTINE:
			Collection_Subroutine(parsing_mode, c, plinda);
			Out_PLinda(c->in_buff, plinda);
			Out_PLinda_Headers(plinda);
			return;
		case IMPLICIT:
			Collection_IMPLICIT(c, plinda);
			break;
		case COMMON:
		case DIMENSION:
			Collection_COMMON(c, plinda);
			break;
		case INTEGER:
		case REAL:
		case DOUBLE:
		case CHARACTER:
			Collection(parsing_mode, c, plinda);
			break;
		case GID:
			CON_Change_GID_to_INT8(c);
			Collection(parsing_mode, c, plinda);
			break;
		case INCLUDE:
			Collection_Include(c, plinda);
			return;
		case PL_OUT:
		case PROC_EVAL:
		case XCOMMIT:
		case XSTART:
		case XRECOVER:
		case ARG_RDP:
		case PL_IN:
		case PL_INP:
		case PL_RD:
		case PL_RDP:
		case CREATE_GROUP:
		case DESTROY_GROUP:
			Emit_Pl_Command(parsing_mode, c, plinda, SymbolTable);
			return;
		case PRINT:
 			Out_PLinda(c->in_buff,plinda);
			return;
		default:   /** default - no action, just echo line **/
			break;
	}

	if (parsing_mode != INCLUDE) 
		Out_PLinda(c->in_buff, plinda);

/***
Print_SymTable(SymbolTable);
***/
}



/*****
Function:	Decorate_Code
Purpose:	Decorate the PLinda code with comments heading each line
		which contains PLinda keyword, so that the file may be compiled
		under the fortran compiler to check if the source code has any
		syntax errors.
Called From:	lex_key_word()
Last Altered:	11.15.95
*****/
void
Decorate_Code(int parsing_mode, CONTEXT * c, FILE * plinda)
{
/**
	fprintf(stderr,"Decorate_Code: Top of => %s\n", c->in_buff);
	fflush(stderr);
**/
	switch (parsing_mode)
	{
		case GID:
		case PROC_EVAL:
		case PL_OUT:
		case PL_IN:
		case PL_INP:
		case PL_RD:
		case PL_RDP:
		case XSTART:
		case XCOMMIT:
		case ARG_RDP:
		case XRECOVER:
		case CREATE_GROUP:
		case DESTROY_GROUP:
			Out_PLinda("C          ",plinda);
			Out_PLinda(c->in_buff, plinda);
	}
}




/****
Function:	lex_key_word
Purpose:	This function is called at the beginning of each line.
		This function serves the purpose of echoing the source code read
		no matter what and recording any information about the variables
		that might be valuable to PLinda.
Note:		Why is the token matching algorithm so horrific in performance ?
			-WhiteSpace doesn't matter in Fortran complicates things...
Misc.:		CONTEXT * c is a buffer of the current line we are reading. 
		Variable 'Mode' denotes the state of parsing the file we are in.
		There are 2 'Modes' : DECORATE, GENERATE.
		DECORATE comments out all PLinda calls from the source code to
		check if the file compiles correctly.
		GENERATE translates to PLinda/Fortran.
Called From:	
Last Alteration:11.26.95
****/
void
lex_key_word(CONTEXT * c, FILE * plinda, int Mode)
{
	int i, j, k, l;
	int backup, parsing_mode;
	int depth_read;
	char test_char, key_char;
	char temp[50] ;
	static struct keyword
	{
		char lexeme[MAX_LEN_KEY];
		int token;
		int length;
	} keyword_table[] =
	{      
		/*keyword*/     	/*Type*/  /*Length of keyword*/		
		{"data", 		DATA,		4},
		{"real", 		REAL,		4},
		{"integer", 		INTEGER,	7},
		{"character", 		CHARACTER,	9},
		{"complex", 		COMPLEX,	7},
		{"doubleprecision", 	DOUBLE,		15},
		{"logical", 		LOGICAL,	7},
		{"common", 		COMMON,		6},
		{"dimension", 		DIMENSION,	9},
		{"implicit", 		IMPLICIT,	8},
		{"program",  		SUBROUTINE, 	7},
		{"subroutine",  	SUBROUTINE, 	10},
		{"include",		INCLUDE,	7},
		{"gid", 		GID,		3},
		{"pl_proc_eval", 	PROC_EVAL,	12},
		{"pl_out", 		PL_OUT,		6},
		{"pl_in", 		PL_IN,		5},
		{"pl_inp", 		PL_INP,		6},
		{"pl_rd", 		PL_RD,		5},
		{"pl_rdp", 		PL_RDP,		6},
		{"xstart", 		XSTART,		6},
		{"xcommit", 		XCOMMIT,	7},
		{"pl_arg_rdp", 		ARG_RDP,	10},
		{"xrecover", 		XRECOVER,	8},
		{"create_group", 	CREATE_GROUP,	12},
		{"destroy_group", 	DESTROY_GROUP,	13},
		{"print", 		PRINT,		5}
	};

 if (Comment(c))		/** If a comment, ignore the line **/  
 {	
	Out_PLinda(c->in_buff, plinda); /* Can turn off if you want. */
	return;
 }	
 while(1)			/* Echo until an End_of_Line is seen.*/
 {
	k = sizeof(keyword_table)/sizeof(struct keyword); /*Step in 'for'*/
	depth_read = 0;	
        for (i=0; i < k; i++)  /*Try to match each entry in keyword table.*/
        {
		backup = 0;
                for (j=0; j < MAX_LEN_KEY; j++)
                {
			do
                        {
                                test_char = CON_getc(c);
				temp[j] = test_char;
				backup++;
				if (End_of_Line(test_char))
				{
					for (j++; j < 50; j++)
						temp[j] = '\0';
					Out_PLinda(c->in_buff, plinda);
					return;
				}
                        }
                        while (White_Space(test_char));

/**
		fprintf(stderr,"\ttest_char => %c ; j => %d\n",test_char, j);	
**/

                        key_char = keyword_table[i].lexeme[j];
			if (!(char_test(key_char, test_char)))
			{
   				CON_restore(c,backup);
				if (j > depth_read)
					depth_read = j;
				for (l=0; l < j; l++)
					temp[l] = '\0';
				break;
			}

                        if ((j+1) == keyword_table[i].length)
			{
				parsing_mode = keyword_table[i].token;
				if ((Mode == DECORATE) && (parsing_mode >= PLINDA))
					Decorate_Code(parsing_mode, c, plinda);
				else
					ParseKeywords(parsing_mode, c, plinda);
				return;
			}
		}
	}

	/** 
	Didn't match any of our keywords. 
		(1) 	Echo the input line out to <plinda.f>	
		(2)	Advance the 'context' input buffer.    
	**/
	for (j=0; j <= depth_read; j++)
	{
		test_char = CON_getc(c);
		if (End_of_Line(test_char))
		{
			Out_PLinda(c->in_buff, plinda);
fprintf(stderr,"Echo out to PLinda => %s\n",c->in_buff);
fflush(stderr);
			return;
		}
	}
 }				/** End while(1) **/
}




/*****
Function:	Comment()
Purpose:	If a comment decl. is passed. Clear the buffer to get to the 
		next line.
Note:		We may want to echo out the comments just to make sure the 
		programs are perfect (via 'diff' func.) - Add this feature 
		later...
*****/
int
Comment(CONTEXT * c)
{
	char test_char;
	test_char = CON_getc(c);
	switch(test_char)
	{
		case 'C':
		case 'c':
		case '*':
			return 1;
		default:
   			CON_restore(c,1);
			return 0;
	}
}



/*****
Function:	Line_Number()	
Called From:	lex_key_word()
Purpose :	Called at the beginning of each line, this function
		checks if the first non-whitespace char. is a number
		(denoting the line_number).  If it is a line number,
		then it is echoed out into the <plinda.f>
Assumption 1: 	We make an assumption that there may no spaces in
		the line number. 
		(ie.	
			Legal		19	i = i + 1
			Illegal		1   9	i = i + 1
		)
Assumption 2: 	Line Numbers can only be 10 decimal digits long.
*****/
void
Line_Number(CONTEXT * c, FILE * plinda)
{
	int i, j;
	char temp[10];

	for (i = 0; i < 10; i++)
		temp[i] = '\0';
	i = 0; 
/**
(1)	Skip over White-Space.
**/
	j = 0;
	while (White_Space(c->in_buff[j]))
		j++;	

/**
(2)	Test to see if decimal digit
(3)	Collect the digit characters.
**/
	while ((c->in_buff[j] >= 48) && (c->in_buff[j] <= 57))
	{
		temp[i] = c->in_buff[j];
		i++;
		j++;
	}
		
	if (i > 0)
		Out_PLinda(temp, plinda);
}

