/* context.c */
/* this is a buffer datatype that allows the following operations
   creation - given a file pointer create a context
   next line - get the next line from the input file
   dispose -   free up memory associated with a context
   get char -  get next non white space character
   get char plain - get next character
   backup - go back to previous character
   set line start - set the line as starting at current position
   reset - reset the cursor to the beginning of the line 
   print - print out the current context	
*/


#include<stdio.h>
#include<stdlib.h>
#include "context.h"
#include "Match_Tools.h"

CONTEXT *CON_create(FILE *f)
{

   CONTEXT *temp;
   temp=(CONTEXT*)malloc(sizeof(CONTEXT));      /* allocate memory */
   temp->fp=f;
   if(fgets(temp->in_buff,BUF_LENGTH,temp->fp)!=NULL) 
   {
     temp->i=0;
     temp->line_start = 0;
     temp->line_number = 1;
     temp->lookahead_eof = 0;	
     temp->lookahead_index = 0;	
     if(fgets(temp->lookahead_buff,BUF_LENGTH,temp->fp)==NULL) 
	temp->lookahead_eof = 1;	
   }
   else 
   {
     free(temp);
     fclose(f);
     fprintf(stderr,"CON_create:: Error, File Pointer passed is NULL\n");
     exit(1);
    }
  return(temp);
}




/*******
Function:	CON_next_line
Purpose:	Get the next line by assigning the input buffer to
		the lookahead buffer(the next line).  Get a new line
		from the file pointer for the lookahead_buffer.
		If the EOF is reached, then lookahead_eof is set to 1. 

Line Cont.:	The index, 'i', for the current buffer is set to
		'lookahead_index' for consideration of skipping over
		the line continuation, '+', when it appears. 
Last Altered:	11.26.95
*******/
int CON_next_line(CONTEXT *c)
{
   int i = 0; 
   c->line_start=0;   /* start the line and cur char off at begin of line */

   if (c->lookahead_eof == 1)
     	return(0);  /* if EOF return 0 */

   for (;i < BUF_LENGTH; i++)
   {	
	c->in_buff[i] = c->lookahead_buff[i];
   }	
   c->i = c->lookahead_index;
   c->lookahead_index = 0;

   if(fgets(c->lookahead_buff,BUF_LENGTH,c->fp)==NULL) 
	c->lookahead_eof = 1;	
   c->line_number++;
   return(1);
}




void CON_dispose(CONTEXT *c)
{
   fclose(c->fp);
   free(c);
}


/******
Function:	CON_getc
Purpose:	Returns the next character in the buffer
******/
char CON_getc(CONTEXT *c)
{
/**  skip white space 
 while(c->in_buff[c->i]==' ' || c->in_buff[c->i] =='\t')
   c->i++;  
**/
if(c->in_buff[c->i]=='\0')
   return('\n');   /* return \n instead of \0 keep \0 in buffer
		      in case wanted to print the buffer out as a string */
else 
 return(c->in_buff[c->i++]); 
}



/*******
Function:	CON_Reset_lookahead
Purpose:	Reset the lookahead_index to 0.
*******/
void 
CON_Reset_lookahead(CONTEXT * c)
{
	c->lookahead_index = 0;
}



/*****
Function:	CON_lookahead_line_cont
Purpose:	Return the 6th column to check if a line continuation
		has been started for the next line.
Altered:	1/28/96
Note:		If there is a tab on the next line, then an assumption
		is made that there is no line continuation. 
		Why??  Because tab lengths can be set to different
		lengths in different editors.  Thus, dangerous!! 
*****/
char CON_lookahead_line_cont(CONTEXT *c)
{
	int j;
	for (j=0; j < 6; j++)
	{
		if (c->lookahead_buff[j] == '\t') 
			return (' ');
	}
	return (c->lookahead_buff[5]);
}


/*****
Function:	Set_lookahead_index
Purpose:	Sets the lookahead index to 6.
		This is done in the event that reading of the next line's
		will begin at the 7th column.  This is done so that the
		6th column's character will not be viewed as part of the 
		first char. string on the next line. 
Altered:	1/30/96
*****/
void
Set_lookahead_index(CONTEXT * c)
{
	c->lookahead_index = 6;
}




char CON_getc_plain(CONTEXT *c)
{
if(c->in_buff[c->i]=='\0')
   return('\n');  
else  
 return(c->in_buff[c->i++]); 
}



/******
Function: 	CON_backup(CONTEXT *c)
Revision Made:	11/1/95
		Changed the backup to not automatically back up on all spaces.
New Author:	Tom Brown
******/
void CON_backup(CONTEXT *c)
{
	if (c->i!=c->line_start)
 		c->i--;  		/* don't back up  past line start */
/***
	while( c->i > c->line_start+1 && (c->in_buff[c->i]==' ' || c->in_buff[c->i] =='\t'))
   		c->i--;
***/
}


/******
Function: 	CON_restore(CONTEXT *c)
Last Change:	11/2/95
		Modular function to call CON_backup to restore the Context
		buffer to the desired position via 'back_spaces'.
Author:		Tom Brown
******/
void 
CON_restore(CONTEXT *c, int back_spaces)
{
	int i;
	for (i = 0;i < back_spaces;i++)
                CON_backup(c);
}


void CON_reset(CONTEXT *c)
{
	c->i=c->line_start;
}

void CON_set_line_start(CONTEXT *c)
{
	c->line_start = c->i;
}



/****
Function:	CON_Beginning_of_Line
Purpose:	Returns 1 if the lexical context search is starting at the 
		beginning of the line.
Why?? :		This function is called from Set_lex_state() at each 
		invocation of next_token() so that LABELS (Line #'s)
		may be effectively stripped out painlessly in the lexer -
		see (strip_label())
****/
int 
CON_Beginning_of_Line(CONTEXT * c)
{
	if (c->line_start == 0)
		return 1;
	return 0;
}



/******
Function:	CON_Change_GID_to_INT8
Purpose:	Changes to GID keyword type declaration in the
		context buffer itself to read INTEGER * 8.
		That is it changes GID to an 8 byte integer. 
Liberties:	Since we have already parsed 'gid' in <Match_keywords.c>,
		we can take liberties with the fact that we know
		it is on that line.
Assumption:	I assume that the keyword 'gid' may not be spelled 
		across many different lines using the Fortran line
		continuation mechanism.
Special Note:	Since the CON_getc() function is used and is based
		off of the integer variable c->i, we need to update
		the variable i to reflect the newly adjusted position
		it should assume for c->in_buff[].
Altered:	2/1/96
******/
void
CON_Change_GID_to_INT8(CONTEXT * c)
{
	char temp[80];
	int index = 0;
	int i = 0;
	int k = 0;
/**
fprintf(stderr,"in_buff => %s\n",c->in_buff);
**/
	while (1)
	{
		if ((c->in_buff[i] == 'G')||(c->in_buff[i] == 'g'))
			break;
		i++;
	}		
	k = i;
	while (1)
	{
		if ((c->in_buff[k] == 'I')||(c->in_buff[k] == 'i'))
			break;
		k++;
	}		
	while (1)
	{
		if ((c->in_buff[k] == 'D')||(c->in_buff[k] == 'd'))
			break;
		k++;
	}		
	k++;
/* Copy from character 0 up to the first character of 'G' or 'g' */ 
	while (index < i) 
	{
		temp[index] = c->in_buff[index];
		index++;
	}
	temp[index++] = 'i';
	temp[index++] = 'n';
	temp[index++] = 't';
	temp[index++] = 'e';
	temp[index++] = 'g';
	temp[index++] = 'e';
	temp[index++] = 'r';
	temp[index++] = ' ';
	temp[index++] = '*';
	temp[index++] = ' ';
	temp[index++] = '2';
/* Set the current char. to be read next in the context buffer */
	c->i = index;

/* Copy the rest of the line into temp[].*/
	while (c->in_buff[k] != '\n')
	{
		temp[index++] = c->in_buff[k++];
		if (c->in_buff[k] == '\0')
			break;
	}
	temp[index] = c->in_buff[k];
	index++;
	temp[index] = '\0';

/* Lastly, copy it back to the c->in_buff[].*/
	i=0;
	while (1)
	{
		c->in_buff[i] = temp[i]; 
		if (temp[i] == '\0')
			break;
		i++;
	}
	c->in_buff[i] = '\0';
/**
fprintf(stderr,"Changed gid in_buff => %s\n",c->in_buff);
**/
}


void
CON_print(CONTEXT * c)
{
/**
	fprintf(stderr,"Current Context => %s\n",c->in_buff);
**/
	fprintf(stderr,"\t\t%s\n",c->in_buff);
}


/******
Function:	CON_Align_to_Char
Purpose:	Do getc function calls until hitting the 'alignment_char'.
******/
void
CON_Align_to_Char(char alignment_char, CONTEXT * c)
{
	char temp;
	CON_reset(c);
	do {
                temp = CON_getc(c);
		if (End_of_Line(temp)) {
                        return;
                }
        } while (temp != alignment_char);
}



