
// This may look like C code, but it is really -*- C++ -*-
//
// File:        ClientProcess.C
// Description: 
// Created:  
// Author:      Karp Joo Jeong
// Mail:        jeong@cs.nyu.edu

#include <stdio.h>
#include <sys/types.h>
#include <assert.h>

// PLinda header files
#include "ErrorType.h"
#include "System_Patterns.h"
#include "ServerCommLink.h"
#include "CheckpointManager.h"
#include "DaemonProcess.h"
#include "ClientProcess.h"
#include "Transaction.h"
#include "TupleGroup.h"
#include "ProcessManager.h"
#include "TransactionManager.h"
#include "TupleGroupManager.h"
#include "ObjectSpace.h"
#include "EventLog.h"
#include "Timer.h"
#include "Predictor.h"

#ifdef _OUTLINE_
#define inline
#include "ClientProcess.iC"
#endif

#ifdef  AUTO_FAULT_MODE
#include "FaultLevelManager.h"
#endif

ClientProcess::ClientProcess(procid my_pid, 
			     procid parent_pid,
			     const physid& arg_id,int interactive) 
#ifdef __GNUG__
: Process(CLIENT, my_pid) , interactive_(interactive)
#else
: Process(PLINDA_CLIENT, my_pid) , interactive_(interactive)
#endif
{
  x_blk_group = NULL;
  x_blk_tuple = NULL;
  x_blk_header = NULL;

  firstTime = true; // for blocking ins and rds
  waitingH = NULL;
  waitingT = NULL;
  x_parent_id = parent_pid;
  x_arg_id = arg_id;
  replayListCommittedLength = 0;
  x_trans = 0;
  switch(ObjectSpace::ft_support) {
  case NO_SUPPORT:
    cur_ft_level = Header::NO_FT;
    break;
  case MESSAGE_REPLAY:
    cur_ft_level = Header::REPLAY_FT;
    break;
  case PRIVATE_SNAPSHOT:
  case ALL:
    cur_ft_level = Header::PRIVATE_FT;
    break;
  case GLOBAL_SNAPSHOT:
    cur_ft_level = Header::GLOBAL_FT;
    break;
  default:
    assert(0);
  }

  //  if(interactive) {
  //    cout << "changing interactive process to message logging replay!!!\n";
  //    cur_ft_level = Header::REPLAY_FT;
  //  }     

  x_temp_trans = 0;
  x_tuple_handle.clear();



  // at the beginning, the server and each process are consistent.
  needLogForCheckpoint_ = false;
  needLogForMigration_ = false;
  needLogForModeChange_ = false;
  outstandingLogRequest_ = false;
  x_consistent = true;
  x_private_log = 0;
  
  // NOTE: chkpt_mgr must already have recovered all the data 
  // when ClientProcess objects are created.
  // log existence flag 
  // directory where to store a snapshot.
#ifdef PETER
   x_cur_log_dir = ObjectSpace::chkpt_mgr.next_chkpt_dir();
  // directory where to store a snapshot.
  x_last_log_dir = ObjectSpace::chkpt_mgr.last_chkpt_dir();
#endif
  x_daemon_proc = 0;
  x_group_list.clear();
  replayList.clear();
  //  outList.clear();
  x_exec_name = NULL;
  x_file = 0;
  x_lineno = 0;

}

ClientProcess::~ClientProcess(void) {
   if(x_file)
     delete [] x_file;
   x_file = 0;
   delete [] x_exec_name;
}


TupleGroup* ClientProcess::find_group(const gid&  g) 
{
   Pix g_cur;
   for(g_cur = x_group_list.first(); g_cur != 0; x_group_list.next(g_cur)) {
      TupleGroup* group = x_group_list(g_cur);
      if(group->get_grpid() == g) {
	 return group;
      }
   }

   TupleGroup* group = ObjectSpace::group_mgr.open_group(g);
   if(group != 0) {
      x_group_list.prepend(group);
   }
   return group;
}


// each tuple group maintains a reference count.
// when opened by a process, it increases its reference count by one;
// it decreases the count by one when closed.

void ClientProcess::close_groups(void) {
   Pix g_cur;
   for(g_cur = x_group_list.first(); g_cur != 0; x_group_list.next(g_cur)) {
      TupleGroup* group = x_group_list(g_cur);
      ObjectSpace::group_mgr.close_group(group);
   }
}

void ClientProcess::stat_end()
{
  Header reply;
  reply.clientRequest(STAT_END);
  reply.transientId(transient_id());
  Tuple *replyTuple = ObjectSpace::stat_collect();
  assert(replyTuple);
  reply.tupleLength(replyTuple->length());
  if(send_reply(reply,replyTuple) == -1)  
    declare_failure();    
  Tuple::destroy(replyTuple);
  replyTuple = NULL;
}
void 
ClientProcess::handle_request(Header *header, Tuple *tuple) {

   Transaction *cur_trans = NULL;
   Tuple::OpType opType;
   TupleGroup *group = NULL;
   char* groupName = NULL;
   Tuple *replyTuple = NULL;
   Tuple *curTuple = NULL;
   Header reply;

   assert(header);
   assert(x_daemon_proc);
   assert(x_daemon_proc->username());

   file(header->fileName());
   lineno(header->lineNum());

   reply.clientRequest(header->clientRequest());
   reply.transientId(transient_id());

   //   cerr << "the request is " << header->clientRequest() << endl << flush;


   if(outstandingLogRequest_ && header->clientRequest() != LOG_FLUSH) {
#ifndef NDEBUG
     if(waitingH ) {
       cerr << "new request = " << header->clientRequest() << endl << flush;
       cerr << "old request = " << waitingH->clientRequest() << endl << flush; 
     }
#endif
     assert(waitingH == NULL);
     assert(waitingT == NULL);
     waitingH = header;
     waitingT = tuple;
     return ;
   }

   if(header->xstart()) {
     //     cerr << "it is " << header->clientRequest() << endl << flush;
     assert(x_trans == 0);
     x_trans = ObjectSpace::trans_mgr.create_transaction(this);
     assert(x_trans != 0);
   }

   switch(header->clientRequest()) {
    case NUM_MACHINES:
      replyTuple = Tuple::create(1,0);
      replyTuple->setActual(0, TupleField::PLint, 
			    ObjectSpace::proc_mgr.num_machines());
      reply.tupleLength(replyTuple->length());
      if(send_reply(reply,replyTuple) == -1) {
	declare_failure();    
	break;
      }
      if(ft_degree() == Header::REPLAY_FT && x_trans) 
	replayList.append(replyTuple);
      else {
	Tuple::destroy(replyTuple);
	replyTuple = NULL;
      }
      break;
   case STAT_BEGIN:
     ObjectSpace::stat_begin();
     break;
   case STAT_END:
     ObjectSpace::stat_end(this);
     break;

   case CHANGE_INTERVAL: {
     int i;
     assert(tuple);
     assert(tuple->numOfFields() == 1);
     tuple->read(0, TupleField::PLint, i);
     ObjectSpace::set_config(ObjectSpace::CHKPT_INTERVAL,(unsigned)i);
     cerr << "Changed the checkpoint interval to " << i << endl << flush;
   }
   break;
   case CHANGE_INTERACTIVE: {
     int i;
     assert(tuple);
     assert(tuple->numOfFields() == 1);
     tuple->read(0, TupleField::PLint, i);
     ObjectSpace::proc_mgr.change_interactivity(identifier(),i) ;    
     cerr << "changed the interactivity of proc " << identifier() <<
       " to " << i << endl << flush;
   }
   break;
   case CHANGE_MODE:
     destroy_log_and_replaylist();
     x_private_log = tuple;
     x_consistent = true;
     // if going into REPLAY, bring the log up to date on the cur tran
     if(header->ftMode() == Header::REPLAY_FT && cur_trans != NULL) {
       // nothing we can do here since we don't know the order
       // of rds and ins!
       cerr << "Cannot change fault tolerance mode to replay in a tran\n" 
	    << flush;

     } else {
       cur_ft_level = header->ftMode();
       cerr << "Process switched to ftmode: " <<  header->ftMode() << endl << 
	 flush;
     }
     break;
   case PRAGMA_TRANS:
     assert(0);
     break;
   case XCOMMIT:
   case TUPLE_CREATIONS: {
     DLList<Tuple*> outList;
     assert(outList.length() == 0);

     cur_trans = x_trans == 0 ? start_temp_trans() : x_trans;

     if(tuple && tuple->opType() == Tuple::FLATTENED)
       unflattenTuples(tuple,outList);
     else if(tuple)
       outList.append(tuple);

     for(Pix p = outList.first() ; p ; outList.next(p)) {
       curTuple = outList(p);
       assert(curTuple);

       if(curTuple->opType() == Tuple::EVAL) {
	 assert(curTuple->group() == zero_gid);
	 cur_trans->eval_tuple(curTuple);
	 curTuple = NULL;
       } else if(curTuple->opType() == Tuple::OUT) {
	 assert(curTuple->group() != zero_gid);
	 TupleGroup *  group = find_group(curTuple->group());
	 assert(group); // later change this to more friendly
	 assert(curTuple->identifier() == zero_physid);
	 x_tuple_handle.set(group, curTuple);
	 cur_trans->access_tuple(group, x_tuple_handle, Tuple::OUT);
	 curTuple = NULL;
       } else {
	 assert(curTuple->opType() == Tuple::XCOMMIT);
	 assert((outList.next(p), p == NULL));
       }
       assert(!curTuple || curTuple->opType() == Tuple::XCOMMIT);
     } // end of for
     outList.clear();
     assert(!curTuple || curTuple->opType() == Tuple::XCOMMIT);
     if(header->clientRequest() == XCOMMIT && 
	ft_degree() == Header::REPLAY_FT) {
       replayList.append((Tuple *const)NULL); // indicator of a commit!
     }
     if(header->clientRequest() == XCOMMIT || x_trans == 0) {
       if(!x_trans)
	 assert(x_temp_trans);
       xcommit(curTuple,header->loop(),header->secs());
       reply.ftMode(cur_ft_level);
     }
     if(send_reply(reply) == -1) {
       declare_failure();    
       break;
     }
   }
   break;			  
   case IN:
   case RD: 
     opType = header->clientRequest() == RD ? Tuple::RD : Tuple::IN;
     cur_trans = x_trans == 0 ? start_temp_trans() : x_trans;
     assert(cur_trans);
     if(x_trans == 0) {
       assert(x_temp_trans);
     }
     x_tuple_handle.clear();
     group = find_group(header->g());
#ifndef NDEBUG
     if(group == 0) {
       cerr << "PLinda user error: trying to in/rd to a " <<
	 "bad tuple handle. restart the computation\n" << flush;
       terminate();
       reply.status(E_INV_GID);
       send_reply(reply);
       declare_error(E_INV_GID);
       return ;
     }
#endif
     group->retrieve_tuple(x_tuple_handle,tuple, opType);
     if(header->synchronyType()==BLOCK &&!x_tuple_handle.valid()) {
       cur_trans->retrieve_tuple(x_tuple_handle,tuple,opType);
       if(!x_tuple_handle.valid()) {
	 Pix cur = group->block_request(this, cur_trans, tuple, opType);
	 status(REQUEST_BLOCKED);
	 x_blk_group = group;
	 x_blk_request = cur;
	 x_blk_tuple = tuple;
	 x_blk_header = header;
	 return ;
       }
     }
     else if(x_tuple_handle.valid()) {
       cur_trans->access_tuple(group, x_tuple_handle, opType);
     }
     if(x_trans == 0) {
       assert(x_temp_trans);
     }
     // main thing that is set up above is the x_tuple_handle!
     
     // make sure the below code is the same in ClientProcess::resume
     // since that is the code that is used when a blocking in or
     // rd resumes!

     reply.tupleLength((x_tuple_handle.valid() ? 
			x_tuple_handle.tuple()->length() : 0));
     
     if(send_reply(reply, x_tuple_handle.tuple()) == -1) {
       declare_failure();
       break;
     }
     if(ft_degree() == Header::REPLAY_FT && x_trans) {
       // could be an empty tuple if inp or rdp
       assert(x_tuple_handle.tuple());
       replayList.append(Tuple::copy(x_tuple_handle.tuple()));
     }
     if(x_trans == 0) {
       xcommit();
     }
     break;
   case CREATE_GROUP:   
      cur_trans = x_trans == 0 ? start_temp_trans() : x_trans;
      assert(tuple);
      tuple->read(0,TupleField::PLchar, groupName);
      group = ObjectSpace::group_mgr.create_group(this, cur_trans, groupName);
      assert(group);
      delete  [] groupName;
      groupName= NULL;
      reply.g(group->get_grpid());
      if(send_reply(reply) == -1) {
	declare_failure();
	break;
      }
      if(ft_degree() == Header::REPLAY_FT && x_trans) {
	Tuple *temp = Tuple::create(0,0);
	temp->group(group->get_grpid());
	temp->opType(Tuple::GROUP);
	replayList.append(temp);
      }
      if(x_trans == 0) {
	assert(x_temp_trans);
	xcommit();
      }
      break;
   case RECORD_READ:
     assert(tuple);
     replayList.append(tuple);
     if(send_reply(reply) == -1) {
       declare_failure();
       break;
     }
     break;
   case ARG_RDP: 
     cur_trans = x_trans == 0 ? start_temp_trans() : x_trans;
     assert(!(x_trans && ft_degree() == Header::REPLAY_FT)); 
     //don't do arg_rdp in a transaction in replay mode
     ObjectSpace::proc_mgr.retrieve_argument(x_tuple_handle, cur_trans,
					     tuple,  x_arg_id);
     reply.tupleLength((x_tuple_handle.valid() ? 
			x_tuple_handle.tuple()->length() : 0));
     if(send_reply(reply,x_tuple_handle.tuple()) == -1) {
       declare_failure();
       break;
     }
     // for now just assume arg_rdp not in a logging transaction
     //     if(ft_degree() == Header::REPLAY_FT && x_trans) {
     //       assert(x_tuple_handle.tuple()); // for now need it to be non null
     //       replayList.append(Tuple::copy(x_tuple_handle.tuple()));
     //     }
     if(x_trans == 0) {
       assert(x_temp_trans);
       xcommit();
     }
     break;
   case XRECOVER:
     {
       if(replayList.length() == 0 && !x_consistent) {
	 // we do not have a message log and we are not
	 // consistent, no way to recover
	 // we must have been in GLOBAL_FT mode and failed
	 // in that mode or while switching to one of the others
	 cerr << "A process is calling xrecover to the server\n";
	 cerr << "But not enough info to recover the guy :(\n";
	 assert(0);
       }
       assert(!x_private_log || x_private_log->opType() == Tuple::XCOMMIT);
       Tuple *temp = flattenTuples(replayList,0,1,x_private_log);
       assert(!temp || temp->opType() == Tuple::FLATTENED || 
	      temp->opType() == Tuple::XCOMMIT);
       reply.tupleLength(temp ? temp->length() : 0);
       if(send_reply(reply, temp) == -1) {
	  declare_failure();
	  break;
	}
       assert(!needLogForMigration_);
       if(needLogForCheckpoint_ || needLogForModeChange_) 
	 request_continuation();
     }
    break;
   case LOG_FLUSH:
     assert(!tuple || tuple->opType() == Tuple::XCOMMIT);
     assert(!tuple || tuple->valid());
     assert(outstandingLogRequest_);
     assert(needLogForCheckpoint_ || needLogForModeChange_ || 
	    needLogForMigration_);

     // the latter case happens when we told the client to switch to
     // commit consistent mode at the last commit,
     // and now it is sending us the log flush as a result!
#ifndef NDEBUG
     if(x_private_log) {
       if(!tuple) // why going from continuation to none??
	 {
	   cerr << "bad going from the continuation " << *x_private_log <<
	     " to none\n" << flush;
	 }
     }
#endif
     Tuple::destroy(x_private_log);
     x_private_log = NULL;
     destroy_replaylist_before_last_commit();
     
     x_private_log = tuple;
     x_consistent = true;
     if(needLogForCheckpoint_) {
       checkpoint_private_log();
       ObjectSpace::chkpt_mgr.continuation_stored(this);
     }

     needLogForCheckpoint_ = false;
     needLogForModeChange_ = false;
     needLogForMigration_ = false;
     outstandingLogRequest_ = false;
     if(waitingH) {
       handle_request(waitingH,waitingT);
       waitingH = NULL;
       waitingT = NULL;
     }
     break;
    case PROC_START:
      // more than one process with the same process identifier may run
      // and try to communicate with the server at the same time.
      // we need to distinguish them.

      transient_id(header->transientId());
      if(x_daemon_proc != 0) 
	ObjectSpace::cleanup_log(x_daemon_proc->hostname(), 
				 x_daemon_proc->username(),transient_id());
      else 
	ObjectSpace::cleanup_log(0, NULL, identifier());
      
      // inform this client whether it restarts after failure.
      if(status() == FAILURE_HANDLED) 
	reply.restartFlag(1);
      status(RUNNING);

      // inform this client of the current fault tolerance level.
      reply.ftMode(ft_degree());
      if(send_reply(reply) == -1) {
	declare_failure();
	break;
      }
      // reset the status.
      break;
    case PROC_END:
      assert(!needLogForCheckpoint_);
      if(send_reply(reply) == -1) {
#ifndef NDEBUG
	cerr << "Client " << identifier() << " died while ending\n";
#endif
      }
      status(COMPLETED);
      comm_link()->set_done();
      // shutdown the link.;
      delete comm_link();
      comm_link(0);
      // clean up the state;
      cleanup_state();
      // remove the tuple in proc_group and the current psnapshot.;
      remove_persistent_identity();
      break;
    default:
      assert(0);
   }
   destroy_current_request(header,tuple);
   
   // informs the daemon that there has been a message; in other words,
   // the processor is still fine.
   if(x_daemon_proc != 0) 
     x_daemon_proc->set_alive_flag(1);
}

void
ClientProcess::destroy_current_request(Header *header,Tuple *tuple) {
   /* now it is ready to take more requests (=> RUNNING status) */
  assert(header);
  if(header->clientRequest() != TUPLE_CREATIONS &&
     header->clientRequest() != XCOMMIT &&
     header->clientRequest() != LOG_FLUSH &&
     header->clientRequest() != RECORD_READ && tuple) { 
    // do not destroy the tuple if it was the case that it was
    // put into tuple space.
    Tuple::destroy(tuple);
  }
  delete header;
  
  if(status() == REQUEST_READY) {
    status(RUNNING);
  }
}

void ClientProcess::destroy_log_and_replaylist() {
  Tuple::destroy(x_private_log);
  x_private_log = NULL;
       
  for(Pix p = replayList.first() ; p ; replayList.next(p)) {
    Tuple *temp = replayList(p);
    Tuple::destroy(temp);
  }
  replayList.clear();
  replayListCommittedLength = 0;
}   

void ClientProcess::destroy_replaylist_before_last_commit() {

  //  Tuple::destroy(x_private_log);
  while(replayListCommittedLength > 0) { 
    Pix p = replayList.first(); 
    assert(p);
    Tuple *temp = replayList(p);
    Tuple::destroy(temp);
    replayList.remove_front();
    replayListCommittedLength--;
  }
}   
#if 0
void ClientProcess::migrate(const char *toMachine) {
  assert(!needLogForMigration_);
  if(x_consistent) {
    // do the migration!
    cleanup_state(); // should close the comm_link
  } else {
    needLogForMigration_ = true;
    // somehow make sure that the process has its state and is not
    // in the process of reconstruction, how to do this?? maybe need
    // some help on the client side
    if(1)
      request_continuation();
  }
}
#endif

int ClientProcess::xcommit(Tuple *continuation,int loop, double secs)
{

  if(x_trans == 0)   {
    assert(continuation == NULL);
    assert(x_temp_trans);
    commit_temp_trans();
    return 1;
  }
   // This is definitely a real transaction committing!!!

  assert(!continuation || continuation->opType()==Tuple::XCOMMIT);

   switch (cur_ft_level) {
   case Header::PRIVATE_FT:
     assert(!needLogForCheckpoint_ && !needLogForModeChange_ &&
	    !outstandingLogRequest_ && !needLogForMigration_);
     destroy_log_and_replaylist();
     assert(!x_private_log || x_private_log->valid());
     x_private_log = continuation;
     assert(x_consistent == true);
     break;
   case Header::NO_FT:
     assert(continuation == NULL);
     break;
   case Header::REPLAY_FT:
     replayListCommittedLength = replayList.length() ;
     assert(continuation == NULL);
     x_consistent = false;
     break; 
  case Header::GLOBAL_FT:
     assert(continuation == NULL);
     x_consistent = false;     
     break;
   default:
     assert(0);
   }

   if(loop) {
     // this is a profiling transaction, make a decision as to what mode to
     // go into!
     int rdsNum, rdsLength,outsNum,outsLength;

     int totContinuationSize = 0;
     int numProcs = 0;
     double alpha = 0;

     // Peter fill in the above variables

     assert(x_trans);
     x_trans->collectTranStats(rdsNum,rdsLength,outsNum,outsLength);
     //     myInfo.set(rdsNum,rdsLength,outsNum,outsLength, secs, c, loop);

     double t_cc = 
       commitConsistentTime(secs, continuation ? continuation->length() : 0, 
			    loop,  alpha);

     double t_ml = 
       messageReplayTime(secs, continuation ? continuation->length() : 0, 
			    loop, ObjectSpace::chkpt_interval, alpha, rdsNum, 
			    rdsLength, outsNum, outsLength);

     double t_tr = 
       triggeredTime(secs, continuation ? continuation->length() : 0, 
		     loop, ObjectSpace::chkpt_interval, alpha, 
		     numProcs, totContinuationSize);

     Header::FT_ModeType newMode;
     if(t_cc <= t_ml && t_cc <= t_tr) {
       newMode = Header::PRIVATE_FT;
       cerr << "pragma tells me to change to private ft\n" << flush;
     } else if(t_ml <= t_cc && t_ml <= t_tr) {
       newMode = Header::REPLAY_FT;
       cerr << "pragma tells me to change to replay ft\n" << flush;
     } else {
       cerr << "pragma tells me to change to global ft\n" << flush;
       newMode = Header::GLOBAL_FT;
     }

     if(ft_degree() != Header::PRIVATE_FT && newMode == Header::PRIVATE_FT) {
       needLogForModeChange_ = true;
       request_continuation();
     } else if(ft_degree()== Header::GLOBAL_FT&&newMode == Header::REPLAY_FT) {
       needLogForModeChange_ = true;
       request_continuation();
     }       
     cur_ft_level = newMode;
   }


   x_trans->commit();
   ObjectSpace::trans_mgr.destroy_transaction(x_trans);
   x_trans = 0;
#ifdef PL_DEBUG
   cerr << identifier() << " switched to " << new_ft_level << " from " << cur_ft_level <<
     endl << flush;
#endif

   return 1;
}

// this function is called by ProcessManager::checkpoint
int
ClientProcess::checkpoint_private_log(void) {
   // already terminated; nothing to do any more.
  //  if(status() == COMPLETED) return 1;

   char name[100];
   ::sprintf(name, "%ld.log", identifier());

   int fd = ::open(name, O_CREAT|O_TRUNC|O_WRONLY, 0644);
   if(fd == -1) return 0;

   // locate the current log directory.
#ifdef PETER
      ObjectSpace::chkpt_mgr.change_dir(dir);
#endif

   char buffer[PL_INT_SZ];
   if(x_private_log) {
     //     cerr << "wrote to disk: " << x_private_log->length() << endl<<flush;
     st_int(x_private_log->length(),buffer);
   }
   else {
     st_int(0,buffer);     
   }

   if(::write(fd, buffer, PL_INT_SZ) != PL_INT_SZ) {
     cerr << "PLinda Error in writing to disk...\n" << flush;
      ::close(fd);
      return 0;
   }
   if(x_private_log) {
     assert(x_private_log->valid());
     if(::write(fd, x_private_log, x_private_log->length()) !=
	x_private_log->length()) {
       cerr << "PLinda Error in writing to disk...\n" << flush;
       ::close(fd);
       return 0;
     }
   }
   
   ::close(fd);
   
   return 1;
}





// this function is called by ProcessManager::rollback.
int 
ClientProcess::restore_private_log(void) {
#ifdef PETER
    x_last_log_dir = ObjectSpace::chkpt_mgr.last_chkpt_dir();
    x_cur_log_dir = ObjectSpace::chkpt_mgr.next_chkpt_dir();

    if(ObjectSpace::chkpt_mgr.chkpt_dir_status() == 0) {
      x_private_log = 0;
      return 1;
    }
  ObjectSpace::chkpt_mgr.change_dir(x_last_log_dir);
#endif
  char name[100];
  ::sprintf(name,"%ld.log",identifier());
   
   int fd;
   if((fd = ::open(name, O_RDONLY)) == -1) {
     cerr << "PLinda internal error could not open file to read log\n";
     assert(0);
     return 1;
   }
   int size;
   if(::read(fd, (char*)&size, PL_INT_SZ) != PL_INT_SZ) return 0;

   if(size == 0) {
     x_private_log = NULL;
   } else {
     char* buffer = new char[size];
     if(::read(fd, buffer, size) != size) {
       cerr << "PLinda ERROR reading from disk\n" << flush;
       delete[] buffer;
       ::close(fd);
       return 0;
     }
     x_private_log = (Tuple*)buffer;
     x_private_log->updatePointers();
     //     cerr << "restoring " << *x_private_log << endl << flush;
     assert(x_private_log->valid());
   }
   ::close(fd);

   return 1;
}


// this function is called by the scheduler at any time.
int 
ClientProcess::request_continuation() {

  // in the following cases, we do not call this function.
  assert(status() != EVALED && status() != DISPATCHED && 
	 status() != COMPLETED);

  //  if(!forFTLevelChange)

  if(status() == FAILURE_DETECTED) { 
    // || status() == FAILURE_HANDLED) {
    return 1;
  }

  if(outstandingLogRequest_ == true) {
    return 1;
  }
  outstandingLogRequest_ = true;

   // send a LOG_REQ request to the client process.
  Header req;
  req.clientRequest(LOG_REQ);
  req.transientId(transient_id());
  //  cerr << "making the log request to the client, yeah!\n " << flush;

   if(comm_link() == 0 || status() == FAILURE_DETECTED) {
      declare_failure();
      return -1;
   }
   if(comm_link()->send(req) == -1) {
      declare_failure();
      return -1;
   }
   return 1;
}


int 
ClientProcess::send_reply(Header &header,const Tuple *tuple) {
   status(RUNNING);

   if(comm_link() == 0 || status() == FAILURE_DETECTED) {
#ifdef PL_DEBUG
      cout << "ClientProcess::send_reply found comm_link = 0 or already "
	<< "  a failure\n" << flush;
#endif
      return -1;
   }
   if(comm_link()->send(header,tuple) == -1)  {
#ifdef PL_DEBUG
      cout << "ClientProcess::send_reply could not use comm link to send" << 
	" a reply\n" << flush;
#endif
      return -1;
   }
   return 0;
}


void
ClientProcess::cleanup_state(void) {

  // clean up the log to the last xcommit!!!!
  if(ft_degree() == Header::REPLAY_FT && x_trans) {
    while(replayList.length() > replayListCommittedLength) {
      Pix p = replayList.last(); 
      assert(p);
      Tuple *temp = replayList(p);
      assert(temp);
      Tuple::destroy(temp);
      replayList.remove_rear();
    }
    Pix p = replayList.last();
    if(p) {
      Tuple *temp = replayList(p);
      assert(temp == NULL); //commit marker of the last transaction!
    }
  }
  outstandingLogRequest_ = false;
  firstTime = true;
  waitingH = NULL;
  waitingT = NULL;
  // abort the active transaction, if any.  --------------------------------
  if(x_trans != 0) {
      x_trans->abort();
      ObjectSpace::trans_mgr.destroy_transaction(x_trans);
      x_trans = 0;
   }
  if(x_temp_trans != 0) 
    abort_temp_trans();

  //  destroy_current_request();
    
  /* inform its daemon to get it off its list */
  if(x_daemon_proc != 0) {
    x_daemon_proc->complete(identifier()); // earlier it was fail(..)
  }
  x_daemon_proc = 0;


  // if blocked, then delete the blocking entry. ----------------------

  if(x_blk_group != 0) {
    x_blk_group->delete_blocked_request(this, x_blk_request);
    x_blk_group = 0;
    x_blk_request = 0;
  }

  /* destroy the commlink obj since it is owned by it */
  if(comm_link() != 0) {
    comm_link()->shutdown();
    delete comm_link();
    comm_link(0);
  }
}


void
ClientProcess::remove_persistent_identity(void) {
   // process termination also requires a sequence of operations.
   // therefore, we perform it as a transaction so that we can
   // later redo it even if we fail in middle of termination.
   {
      Transaction* end_trans;
      end_trans = ObjectSpace::trans_mgr.create_transaction(this);

      // update proc_group and arg_group.
      ObjectSpace::proc_mgr.terminate_PROCESS(end_trans, this);
      end_trans->commit();
      ObjectSpace::trans_mgr.destroy_transaction(end_trans);
   }
   destroy_log_and_replaylist();
   // decrement reference counts of groups.
   close_groups();

}


void
ClientProcess::declare_error(ErrorType err) {
   // set the error type  
   error(err);
   status(ERROR_FOUND);
#ifdef PL_DEBUG
   cerr << "ClientProcess error found: " << err.message() << endl <<  flush;
#endif
   // shutdown the communication link.
   if(comm_link() != 0) {
      comm_link()->set_done(); 
      comm_link()->shutdown();
      delete comm_link();
      comm_link(0);
   }
   cleanup_state();
}


// this process won't continue and so remove info about it.
void
ClientProcess::handle_error(void) {
   // clean up state work. 
   // daemon is also informed in this call...
   cleanup_state();

   // remove persistent information about this process.
   remove_persistent_identity();

   // Don't tell the checkpoint manager about this processes continuation
   // better to lock up the checkpointing algorithm so the user
   // can restore from an old one!

   // in any case, we consider this as terminated.
   status(COMPLETED);

}


void
ClientProcess::declare_failure(void) {
   // already found or handled.
   // assert(!failure_handled() && !failure_detected());

   // if fault tolerance is disabled, then failure is treated as error.
   // therefore, no reaction is taken.

   if(ft_degree() == Header::NO_FT) {
      declare_error(E_FAIL);
      return;
   }
   status(FAILURE_DETECTED);
   if(comm_link() != 0) {
      comm_link()->shutdown();
      delete comm_link();
      comm_link(0);
   }
   cleanup_state();
}


void
ClientProcess::handle_failure(void) {
   // already handled.
   assert(!failure_handled());

   if(replayList.length() == 0 && !x_consistent) {
     // no way to recover
     ObjectSpace::state(ObjectSpace::NEEDS_REBOOT);
     declare_error(E_NEED_REBOOT);
     return;
   }

   // clean up the intermediate work. 
   //   cleanup_state(); already done in declare failure!
   assert(x_daemon_proc == 0);

   // set the status. 
   status(FAILURE_HANDLED);

   assert(comm_link() == 0);

   // shutdown the communication link. 
   //   if(comm_link() != 0) {
   //      delete comm_link();
   //      comm_link(0);
   //   }


   // prepare for respawning 

   assert(x_trans == 0);

   // respawn this process; move tuple from proc group to eval group.
   assert(replayList.length() != 0 || x_consistent);
   ObjectSpace::proc_mgr.respawn_PROCESS(identifier());

}


int
ClientProcess::check_failure(void) {
   if(comm_link() == 0) {
      return 0;
   }
   return comm_link()->check_failure();
}



void
ClientProcess::kill(void) {
   extern const char *REMOTESH;
   if(x_daemon_proc != 0) {
      char command[1000];
      assert(x_daemon_proc);
      assert(x_daemon_proc->username());
      ::sprintf(command, "%s %s -l %s kill -9 %ld", REMOTESH ,
		x_daemon_proc->hostname(),x_daemon_proc->username(), 
		transient_id());
      cerr << command << "\n";
      ::system(command);
   }
}


void
ClientProcess::terminate(void) {
   if(completed()) return;


   // shutdown the link.
   status(COMPLETED);
   if(comm_link() != 0) {
      comm_link()->set_done();
      delete comm_link();
      comm_link(0);
   }

   // kill the running process
   kill();
   // make sure to cleanup_state after kill so we know the username!
   // clean up the local state. 
   cleanup_state();

   // remove persistent info
   remove_persistent_identity();


   // inform the global snapshot manager.
   if(needLogForCheckpoint_) {
     cerr << "Process being terminated but there was an outstanding " <<
       "continuation request may leave things inconsistent***\n";
     ObjectSpace::chkpt_mgr.continuation_stored(this);
   }

}



// -------------------------------------------------------------------------

String
ClientProcess::print(void) const {
   String hold;
   hold = "client(\n";
   hold += Process::print();

   hold += ",daemon:";
   if(x_daemon_proc != 0) {
      hold += dec(x_daemon_proc->identifier());
   } else {
      hold += "0";
   }
   
   hold += ",trans:";
   if(x_trans != 0) {
      hold += (x_trans->identifier()).print();
   } else {
      hold += "zero_transid";
   } 

   hold += ",request:\n";
   hold += "?";


   hold += ",\n argument:";
   hold += x_arg_id.print();
   
   hold += ")";

   return hold;
}

