// This may look like C code, but it is really -*- C++ -*-
//
// File:     Driver.C
// Created:  
// Author:   Karp Joo Jeong
// Mail:     jeong@cs.nyu.edu
//

#include <stream.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#ifndef linux
#include <sgtty.h>
#endif
#include <ctype.h>

#include "C_Interface.h"
#include "plinda_ids.h"
#include "ErrorType.h"
#include "EventLog.h"

#include "Transaction.h"
#include "Process.h"
#include "ClientProcess.h"
#include "ProcessManager.h"
#include "TransactionManager.h"
#include "TupleGroupManager.h"
#include "Scheduler.h"
#include "ObjectSpace.h"
#include "CheckpointManager.h"
#include "System_Patterns.h"
#include "DaemonProcess.h"
#include "ConnectionListener.h"
#include "MonitorProcess.h"
#include "Architectures.h"
#include "plinda_daemon.h"

// forward declarations ----------------------------------------------------

int get_server_options(int argc, char** argv, char** env, char* main_program,
		       int& recover_system);

int spawn_main_process(const char* name);
int spawn_monitor_process(const char* name);
int dispatch_daemons(const char* host_file);

const int PLINDA_CLIENT_PORT = 6000;

// ---------------------------------------------------------------------------
// main function    
// ---------------------------------------------------------------------------

main(int argc, char **argv, char**env){
  // options
  int 		recover_system;
  char		main_program[100];

  // read the command line arguments   -------------------------------
  if(!::get_server_options(argc, argv, env, main_program,
			   recover_system)) {
    ObjectSpace::event_log.fatal("main", 
				 "unable to parse the command arguments");
  }


  // change the file creation mask so that everyone can access
  // what we make
  ::umask(0011);

    

  // open the event log file ------------------------------------------

  if(recover_system) 
    ::system(form("/bin/mv %s %s~", event_log_file_name, event_log_file_name));


  ObjectSpace::event_log.open(event_log_file_name);
  if(!ObjectSpace::event_log) {
    ObjectSpace::event_log.fatal("main", "unable to create an event log file");
  }

  cout << "The PLinda runtime system is starting up....\n" <<  flush;
  // initialize the server state. 
  if(!ObjectSpace::initialize(recover_system)) {
    ObjectSpace::event_log.fatal("main", 
				 "unable to initialize the server state");
  }


  {
    char command[1000];
    ::sprintf(command,"/bin/mv %s/plinda.cleanup %s/plinda.cleanup~",
	      ObjectSpace::working_directory, ObjectSpace::working_directory);
    ::system(command);
    char uname[64];
    ::cuserid(uname);
    if(!uname) {
      cerr << "Error: system command getlogin returning NULL\n";
      assert(uname);
    }
    ObjectSpace::cleanup_log(ObjectSpace::conn_lsner.hostname(),
			     uname, ::getpid());
  }


  // recover the server, if requested          ---------------------------
  if(recover_system) {
    if(!ObjectSpace::recover()) {
      ObjectSpace::event_log.error("main", 
				   "unable to recover the server state");
      ObjectSpace::event_log.messg("main", "restart from the scratch");

      // recover_system flag is no longer effective. 
      recover_system = 0;
    }
  } 


  if(ObjectSpace::batch_flag == 1) {
    // This is no user interface mode!!!

    spawn_main_process(main_program);

    // the server spawns daemons in a batch mode.
    if(dispatch_daemons(ObjectSpace::host_file) == 0) 
      exit(-1);
  }

  ObjectSpace::scheduler.event_loop();

    // terminate processes
    ObjectSpace::proc_mgr.terminate();
    // before termination, server checkpoints the state. ---------------
    if(!ObjectSpace::chkpt_mgr.close()) {
      ObjectSpace::event_log.fatal("main", 
				   "unable to checkpoint state upon termination");
    }
  {
    // remove log files
    char command[1000];
    ::sprintf(command,"/bin/mv %s/plinda.server %s/plinda.server~",
	      ObjectSpace::working_directory, ObjectSpace::working_directory);
    ::system(command);
    ::sprintf(command,"/bin/mv %s/plinda.cleanup %s/plinda.cleanup~",
	      ObjectSpace::working_directory, ObjectSpace::working_directory);
    ::system(command);
  }
  ObjectSpace::event_log.exit(0); 
}


/* -------------------------------------------------------------------

int
run_as_background(void) {
  // split off the daemon
  cout << "  server is starting as a background process ....\n";

  // clone off its copy process.
  {
    int ret;
    if((ret = fork()) == -1) {
      cerr << "> **ERROR: fork failed\n";
      return 0;
    }

    if(ret) {
      // foreground goes away.
      ::exit(0);
    }
  } 

  // disassociate from the terminal
  {
    // change process group
    ::setpgrp(0, ::getpid());

    // abandon control tty
    int f = ::open("/dev/tty", O_RDWR);
    if(f < 0) {
      cerr << "> **ERROR: unable to open /dev/tty\n";
      return 0;
    }
    ::ioctl(f, TIOCNOTTY, (char*)0);
    ::close(f);
  }
 
  return 1;
}

  ------------------------------------------------------------------- */
void usage() 
{
  cerr << "usage: plinda_server " ;
  cerr << "       [-b main-prgm] ; batch processing mode\n";  
  cerr << "       [-r]           ; recover from server failure\n";  
  cerr << "       [-j dir]       ; server working directory\n";
  cerr << "       [-x x-host]    ; X server host name\n";
  cerr << "       [-w host-file] ; file for workstation names\n";
  cerr << "       [-c interval]  ; checkpoint interval\n";
  cerr << "       [-t interval]  ; failure detection interval\n";
  cerr << "       [-i]           ; minimum idle time for a daemon\n";
  cerr << "       [-f fault-mode]; 1 = " << FtTypeName[1] << "\n";
  cerr << "                        2 = " << FtTypeName[2] << "\n";
  cerr << "                        3 = " << FtTypeName[3] << "\n";
  cerr << "                        4 = " << FtTypeName[4] << "\n";
  cerr << "                        5 = " << FtTypeName[5] << "\n";
  exit(0);
}

int
get_server_options(int argc, char** argv, char** , char* main_program,
		   int& recover_system) {

  extern char* optarg;

  // initialization
  recover_system = 0;


  // parse the command line arguments

  if(argc == 1) // there are always some
    usage();
  ObjectSpace::ft_support = PRIVATE_SNAPSHOT;

  int option_char;
  while((option_char=getopt(argc,argv,"ab:c:hi:j:rt:f:w:x:")) != -1)
  {
    switch(option_char) {
    case 'b':
      ObjectSpace::batch_flag = 1;
      ::strcpy(main_program,optarg);
      break;
    case 'c':
       ObjectSpace::chkpt_interval = atoi(optarg);
       break;
    case 'h':
      usage();

    case 'i':
      ObjectSpace::minimum_idletime = atoi(optarg);
      break;
    case 'j':
      {
	char* ptr = new char[::strlen(optarg)+1];
	::strcpy(ptr,optarg);
	ObjectSpace::working_directory = ptr;
      }
    break;
    case 'f':
      ObjectSpace::ft_support = (FT_LevelType)atoi(optarg);
      assert(atoi(optarg) > 0 && atoi(optarg) < 6);
      cerr << "Server being started in " << FtTypeName[atoi(optarg)] << endl
	   << flush;
      break;
    case 'r':
      recover_system = 1;
      break;
    case 't':
      ObjectSpace::process_timeout = atoi(optarg);
      break;

    case 'w':
      {
	char* ptr = new char[::strlen(optarg) + 1];
	::strcpy(ptr, optarg);
	ObjectSpace::host_file = ptr;
      }
      break;
    case 'x':
      {
	char* ptr = new char[::strlen(optarg)+1];
	::strcpy(ptr, optarg);
	ObjectSpace::x_server_name = ptr;
      }	
    break;
  default:
    cerr << "Bad option to the server!!!! < " << option_char << ">" << "\n";
    assert(0);
    }
  }
  
  
  if(recover_system == 1) {
    cerr << "> recover from server failure.\n"; 
  }
  
    
  if(ObjectSpace::working_directory != 0) {
    if(::chdir(ObjectSpace::working_directory) == -1) {
      cerr << "> can't change the working directory to " 
	   << ObjectSpace::working_directory << "\n";
      cerr.flush();
      ::exit(1);
    }
  }
  
  if(ObjectSpace::x_server_name == 0) {
    cerr << "> warning: no x server name given; \n";
    cerr << "           user invocation required for the main process.\n";
  } 

  cerr << "> checkpoint every " << ObjectSpace::chkpt_interval << 
    " seconds.\n";


  return 1;
}


// -------------------------------------------------------------------
#ifdef linux
static const char *server_type= "linux";
#endif
#ifdef alpha
static const char *server_type= "alpha";
#endif
#ifdef solaris
static const char *server_type="solaris";
#endif
#ifdef sunos
static const char *server_type="sunos";
#endif
#ifdef irix
static const char *server_type= "irix";
#endif

int
dispatch_daemons(const char* host_file) {
   extern const char *REMOTESH;
  if(!ObjectSpace::chkpt_mgr.change_dir(CheckpointManager::ROOT_DIR)) {
    ObjectSpace::event_log.error("dispatch_daemons",
			          "unable to change to the working directory");
    return 0; 
  }

  // open the given host file.
  ifstream hosts(host_file, ios::in);
  if(!hosts) {
    cerr << "> **ERROR: failed to open the host file.\n";
    ObjectSpace::event_log.error("dispatch_daemons", 
				 "unable to open host file");
    return 0;
  }


  int num_hosts = 0; 
  while(1) {
      // read a host name;
      char host_name[1000]; // 1000 must be enough.;
      hosts >> host_name;
      if(!hosts) {
	  break;
      }
      char host_type[1000];
      hosts >> host_type;
      char user_name[1000];
      hosts >> user_name;
      assert(::strlen(user_name) > 0 && ::strlen(user_name) < 20);
      for(int i = 0 ; host_type[i] ; i++) 
	host_type[i] = tolower(host_type[i]);

      if(::strlen(user_name) < 1) {
	cerr << "Invalid host file must have user name in there\n";
	::exit(1);
      }
   if(! PLinda_Architectures::supported(host_type)) {
	 cerr << host_type << " not supported yet, Sorry" << 
	   __LINE__ << __FILE__ << endl << ::flush;
	 assert(0);
      }
   cerr << "> host-add: " << host_name << "\n";
      
      // dispatch a daemon;

      DaemonProcess* daemon = 
	ObjectSpace::proc_mgr.create_daemon(host_name,host_type,user_name);
      if(daemon == 0) {
	ObjectSpace::event_log.error("dispatch_daemons", 
				     "unable to dispatch a daemon");
        return 0;
      }

      char command[1000];
      const int theMode = MONITOR;
      // for now can only run the daemon in normal mode w/no interface
      ::sprintf(command,
		"%s %s -l %s \"~/plinda/bin/%s/plinda_daemon %ld %d %s %s %s %s %d %d\"&", 
		REMOTESH,
		host_name, user_name,host_type,daemon->identifier(), 
		htons(ObjectSpace::conn_lsner.port_number()),
		ObjectSpace::conn_lsner.hostname(),
		(ObjectSpace::x_server_name != 0) ?
		ObjectSpace::x_server_name : "\\?",
		host_type,server_type, theMode,0);

      cout << command << "\n" << flush;
      
      if(system(command) != 0) {
	cerr << "> **ERROR: failed to invoke a daemon on " << host_name 
	     << "\n" << flush;
	ObjectSpace::proc_mgr.destroy_daemon(daemon);
	continue;
      }

      ++num_hosts;
  }

  if(num_hosts == 0) {
    cerr << "> **ERROR: no machine name in the host file.\n";
    exit(1);
  }

  return 1;
}


// ----------------------------------------------------------------------
// insert an eval tuple for the main program 
// and then a daemon will dispatch a process for the tuple.
// of course, do it in a transactional manner. 
int
spawn_main_process(const char* name) {
  
  Tuple* main_prgm_tuple = ::create_main_prgm_tuple(name);

  ObjectSpace::proc_mgr.spawn_PROCESS(ObjectSpace::proc_mgr.super, 
  			              main_prgm_tuple, 0);
  
  return 1;
}


